# -*- coding: utf-8 -*-

import dataclasses

import pytest

import json
from datetime import datetime, timezone

import sayt.api as sayt
from aws_resource_search.documents.resource_document import (
    get_utc_now,
    to_human_readable_elapsed,
    to_utc_dt,
    to_simple_dt_fmt,
    to_iso_dt_fmt,
    get_none_or_default,
    get_description,
    get_datetime_iso_fmt,
    get_datetime_simple_fmt,
    ResourceDocument,
)


def test_human_readable_elapsed():
    assert to_human_readable_elapsed(15) == "15 sec"
    assert to_human_readable_elapsed(60) == "1 min"
    assert to_human_readable_elapsed(90) == "1 min 30 sec"
    assert to_human_readable_elapsed(3600) == "1 hour"
    assert to_human_readable_elapsed(3725) == "1 hour 2.1 min"
    assert to_human_readable_elapsed(86400) == "1 day"
    assert to_human_readable_elapsed(93900) == "1 day 2.1 hour"


def test_to_utc_dt():
    assert to_utc_dt(datetime.utcnow()).tzinfo == timezone.utc


def test_to_simple_dt_fmt():
    assert "+" not in to_simple_dt_fmt(get_utc_now())


def test_to_iso_dt_fmt():
    assert "+00:00" in to_iso_dt_fmt(get_utc_now())


def test_get_none_or_default():
    data = {"a": 1, "b": {"c": 3}}
    assert get_none_or_default(data, "a") == 1
    assert get_none_or_default(data, "d") is None
    assert get_none_or_default(data, "d", "hello") == "hello"
    assert get_none_or_default(data, "b.c") == 3
    assert get_none_or_default(data, "b.d") is None
    assert get_none_or_default(data, "b.d", "hello") == "hello"
    assert get_none_or_default(data, "c.e") is None
    assert get_none_or_default(data, "c.e", "hello") == "hello"


def test_get_datetime_simple_fmt():
    assert get_datetime_simple_fmt({}, "time") == "No datetime"
    data = {"time": datetime(2000, 1, 1)}
    assert get_datetime_simple_fmt(data, "time") == "2000-01-01 00:00:00"


def test_get_datetime_iso_fmt():
    assert get_datetime_iso_fmt({}, "time") == "No datetime"
    data = {"time": datetime(2000, 1, 1)}
    assert get_datetime_iso_fmt(data, "time") == "2000-01-01T00:00:00+00:00"


@dataclasses.dataclass
class S3Bucket(ResourceDocument):
    @classmethod
    def from_resource(
        cls,
        resource,
        bsm,
        boto_kwargs,
    ):
        return cls(
            raw_data=resource,
            id=resource["Name"],
            name=resource["Name"],
        )

    @property
    def title(self) -> str:
        return f"bucket = {self.name}"

    @property
    def arn(self) -> str:
        return f"arn:aws:s3:::{self.name}"

    def get_console_url(self, console) -> str:
        return console.s3.get_console_url(bucket=self.name)


class TestResourceDocument:
    def _test_properties_methods(self):
        doc = ResourceDocument(raw_data={}, id="test", name="test")
        with pytest.raises(NotImplementedError):
            _ = doc.title
        _ = doc.subtitle
        _ = doc.short_subtitle
        _ = doc.autocomplete
        with pytest.raises(NotImplementedError):
            _ = doc.arn
        with pytest.raises(NotImplementedError):
            _ = doc.get_console_url(console=None)
        # with pytest.raises(NotImplementedError):
        #     _ = doc.get_details(ars=None)
        # assert doc.get_initial_detail_items(ars=None) == []

        s3_bucket = S3Bucket.from_resource(
            resource={"Name": "test-bucket", "CreationDate": datetime(2021, 1, 1)},
            bsm=None,
            boto_kwargs=None,
        )
        _ = s3_bucket.title
        _ = s3_bucket.subtitle
        _ = s3_bucket.short_subtitle
        _ = s3_bucket.autocomplete
        _ = s3_bucket.uid
        _ = s3_bucket.arn

    def _test_one_line(self):
        assert ResourceDocument.one_line("hello") == "hello"
        assert ResourceDocument.one_line("hello\nword") == "hello ..."
        assert (
            ResourceDocument.one_line(json.dumps(dict(a=1, b=2), indent=4))
            == '{"a": 1, "b": 2}'
        )
        assert ResourceDocument.one_line("") == "NA"
        assert ResourceDocument.one_line(None, "Not available") == "Not available"
        assert ResourceDocument.one_line(dict(a=1, b=2)) == '{"a": 1, "b": 2}'

    def _test_get_dataset_fields(self):
        @dataclasses.dataclass
        class DummyS3Bucket(ResourceDocument):
            arn: str = dataclasses.field(
                metadata={"field": sayt.StoredField(name="arn")}
            )

        search_fields = DummyS3Bucket.get_dataset_fields()
        assert search_fields[4].name == "arn"
        assert isinstance(search_fields[4], sayt.StoredField)

        @dataclasses.dataclass
        class WrongResource1(ResourceDocument):
            arn: str = dataclasses.field()

        with pytest.raises(KeyError):
            WrongResource1.get_dataset_fields()

        @dataclasses.dataclass
        class WrongResource2(ResourceDocument):
            arn: str = dataclasses.field(metadata={"field": 123})

        with pytest.raises(TypeError):
            WrongResource2.get_dataset_fields()

        @dataclasses.dataclass
        class WrongResource3(ResourceDocument):
            arn: str = dataclasses.field(
                metadata={"field": sayt.StoredField(name="bucket_arn")}
            )

        with pytest.raises(ValueError):
            WrongResource3.get_dataset_fields()

    def test(self):
        self._test_properties_methods()
        self._test_one_line()
        self._test_get_dataset_fields()


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(
        __file__,
        "aws_resource_search.documents.resource_document",
        preview=False,
    )



# -*- coding: utf-8 -*-

"""
In object-oriented programming, the inheritance hierarchy is a pattern where
child objects inherit attributes and methods from parent objects. Similarly,
in configuration management, the global configuration often acts as the
default value, allowing for the possibility of overriding specific values in
environment-specific configurations.

This module implements a simple dialect of JSON that supports inheritance for
better reusability of configuration data. For example, consider the following
config data::

    >>> {
    ...     "_shared": {
    ...         "*.username": "root",
    ...     },
    ...     "server1": {
    ...         "password": "password1",
    ...     },
    ...     "server2": {
    ...         "password": "password2",
    ...     }
    ... }

It will be transformed into::

    >>> {
    ...     "server1": {
    ...         "username": "root",
    ...         "password": "password1",
    ...     },
    ...     "server2": {
    ...         "username": "root",
    ...         "password": "password2",
    ...     },
    ... }

The ``_shared`` key is used to specify the shared values. It is a key value
pair where the key is a path to the value.
"""

import typing as T

SHARED = "_shared"

_error_tpl = (
    "node at JSON path {_prefix!r} is not a dict or list of dict! "
    "cannot set node value '{prefix}.{key}' = ...!"
)


def make_type_error(prefix: str, key: str) -> TypeError:
    if prefix == "":
        _prefix = "."
    else:
        _prefix = prefix
    return TypeError(_error_tpl.format(_prefix=_prefix, prefix=prefix, key=key))


def inherit_shared_value(
    path: str,
    value: T.Any,
    data: T.Union[T.Dict[str, T.Any], T.List[T.Dict[str, T.Any]]],
    _prefix: T.Optional[str] = None,
):
    """
    Try to inherit a shared value to a specific item in a list or dict. If the
    value is already set, then do nothing. It will update the data inplace.

    :param path: JSON path in dot notation to the value to be set.
        Valid path examples: ``key1.key2``, ``databases.*.username``.
        It cannot ends with ``.*``.
    :param value: the value to be set.
    :param data: the data to be updated.
    """
    if path.endswith("*"):
        raise ValueError("json path cannot ends with *!")
    if _prefix is None:
        _prefix = ""

    parts = path.split(".")

    if len(parts) == 1:
        if isinstance(data, dict):
            data.setdefault(parts[0], value)
        elif isinstance(data, list):
            for item in data:
                if not isinstance(item, dict):
                    raise make_type_error(_prefix, parts[0])
                item.setdefault(parts[0], value)
        else:
            raise make_type_error(_prefix, parts[0])
        return

    key = parts[0]
    if key == "*":
        for k, v in data.items():
            if k != SHARED:
                inherit_shared_value(
                    path=".".join(parts[1:]),
                    value=value,
                    data=v,
                    _prefix=f"{_prefix}.{key}",
                )
    else:
        if isinstance(data, dict):
            inherit_shared_value(
                path=".".join(parts[1:]),
                value=value,
                data=data[key],
                _prefix=f"{_prefix}.{key}",
            )
        elif isinstance(data, list):
            for item in data:
                inherit_shared_value(
                    path=".".join(parts[1:]),
                    value=value,
                    data=item[key],
                    _prefix=f"{_prefix}.{key}",
                )
        else:
            raise make_type_error(_prefix, key)


def apply_shared_value(data: dict):
    """
    Transform the data inplace by applying the shared values. It uses
    deep-first search to traverse the data, because deeper node value may
    override the value of the same key in the upper node.
    """
    # implement recursion pattern
    for key, value in data.items():
        if key == SHARED:
            continue
        if isinstance(value, dict):
            apply_shared_value(value)
        if isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    apply_shared_value(item)

    # try to set shared value
    has_shared = SHARED in data
    if has_shared is False:
        return

    # pop the shared data, it is not needed in the final result
    shared_data = data.pop(SHARED)
    for path, value in shared_data.items():
        inherit_shared_value(path=path, value=value, data=data)



# -*- coding: utf-8 -*-

import typing as T
import json

from .utils import guid, envs, rand_env

if T.TYPE_CHECKING:
    from .main import FakeAws


class IamMixin:
    iam_groups: T.List[str] = list()
    iam_users: T.List[str] = list()
    iam_roles: T.List[str] = list()
    iam_policies: T.List[str] = list()

    @classmethod
    def create_iam(cls: T.Type["FakeAws"]):
        for ith in range(1, 1 + 10):
            env = rand_env()

            name = f"{env}-{guid}-{ith}-iam-group"
            cls.bsm.iam_client.create_group(GroupName=name)
            cls.iam_groups.append(name)

            name = f"{env}-{guid}-{ith}-iam-user"
            cls.bsm.iam_client.create_user(UserName=name)
            cls.iam_users.append(name)

            name = f"{env}-{guid}-{ith}-iam-role"
            cls.bsm.iam_client.create_role(
                RoleName=name,
                AssumeRolePolicyDocument="{}",
            )
            cls.iam_roles.append(name)

            name = f"{env}-{guid}-{ith}-iam-policy"
            cls.bsm.iam_client.create_policy(
                PolicyName=name,
                PolicyDocument=json.dumps(
                    {
                        "Version": "2012-10-17",
                        "Statement": [
                            {"Effect": "Allow", "Action": "*", "Resource": "*"}
                        ],
                    }
                ),
            )
            cls.iam_policies.append(name)



# -*- coding: utf-8 -*-

from aws_resource_search.items.detail_item import DetailItem, ExceptionItem


class TestDetailItem:
    def test_new(self):
        item = DetailItem.new(
            title="my title",
            subtitle="my subtitle",
        )
        item.enter_handler(ui=None)
        item.ctrl_a_handler(ui=None)
        assert item.variables.get("copy") is None
        assert item.variables.get("url") is None

        item = DetailItem.new(
            title="my title",
            autocomplete="autocomplete",
        )
        assert "Tab" in item.subtitle

        item = DetailItem.new(
            title="my title",
        )
        assert item.subtitle == "No subtitle"

    def test_from_detail(self):
        item = DetailItem.from_detail(
            key="key",
            value="value",
        )
        assert item.variables["copy"] == "value"
        assert item.variables["url"] is None

        item = DetailItem.from_detail(
            key="key",
            value="value",
            url="url",
        )
        assert item.variables["copy"] == "value"
        assert item.variables["url"] == "url"

    def test_from_env_vars(self):
        item_list = DetailItem.from_env_vars(
            env_vars={"key": "value"},
            url="url",
        )
        assert len(item_list) == 1
        item_list = DetailItem.from_env_vars(
            env_vars={},
            url="url",
        )
        assert len(item_list) == 1

    def test_from_tags(self):
        item_list = DetailItem.from_tags(
            tags={"key": "value"},
            url="url",
        )
        assert len(item_list) == 1
        item_list = DetailItem.from_tags(tags={}, url="url")
        assert len(item_list) == 1

    def test_get_initial_detail_items(self):
        pass

    def test_error_handling(self):
        detail_items = []
        with DetailItem.error_handling(detail_items):
            raise ValueError("something wrong")
        assert isinstance(detail_items[0], ExceptionItem)


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(
        __file__,
        "aws_resource_search.items.detail_item",
        preview=False,
    )



# -*- coding: utf-8 -*-

"""
todo: docstring
"""

import typing as T
import fire

from awscli_mate.ui import run_ui as run_awscli_mate_ui

from .._version import __version__


class ArsCli:
    """
    AWS Resource Search CLI.

    Search AWS resource in terminal, just like you are using AWS Console in web browser.

    You can do things like:

    - open aws console url
    - copy information to clipboard
    - view aws resource details in terminal

    For full documentation, please view https://github.com/MacHu-GWU/aws_resource_search-project

    Usage:

    - ``ars``: enter the AWS resource search interactive UI
    - ``ars -v``: show the version of this tool
    - ``ars -h``: show this help message
    - ``ars ${{sub_command}}``: run the given sub command. See "COMMANDS" section below.
    """

    def __call__(self, version: T.Optional[bool] = None):
        """
        Serve for the ``ars`` command without any arguments.
        """
        from ..ars_def import validate_bsm
        from ..ui_init import ui, run_ui as run_ars_ui

        if version:
            print(__version__)
        else:
            validate_bsm(ui.ars.bsm)
            run_ars_ui()

    def which(self):
        """
        Show AWS account id, alias, region of the current boto3 session.

        Usage:

        ``ars which``, print the AWS account id, alias, region of the current boto3 session.
        Example::

            $ ars which
            AWS Account ID = 123456789012
            AWS Account Alias = my-aws-account-alias
            AWS Region = us-east-1
        """
        from . import which

        which.main()

    def set_profile(self):
        """
        Set one of your AWS named profile as default profile for ARS UI.

        Usage:

        - ``ars set-profile``, then follow the interactive UI to set the profile.
        - Example: https://github.com/MacHu-GWU/awscli_mate-project#use-awscli_mate-as-a-interactive-cli
        """
        run_awscli_mate_ui()

    def clear(self):
        """
        Clear all index and cache of this App.

        Usage:

        - ``ars clear``: clear all index and cache of this App.
        """
        from . import clear

        clear.main()


def run():
    """
    The entry point of this CLI tool.
    """
    ars_cli = ArsCli()
    fire.Fire(ars_cli)



# -*- coding: utf-8 -*-

import botocore.exceptions

from ..ars_init import bsm


def main():
    print(f"AWS Account ID = {bsm.aws_account_id}")
    try:
        res = bsm.iam_client.list_account_aliases()
        aliases = res.get("AccountAliases", [])
        if aliases:
            print(f"AWS Account Alias = {aliases[0]}")
    except botocore.exceptions.ClientError as e:
        pass
    print(f"AWS Region = {bsm.aws_region}")



# -*- coding: utf-8 -*-

"""

"""

import typing as T
import dataclasses
from pathlib import Path

import zelfred.api as zf

try:
    import pyperclip
except ImportError:  # pragma: no cover
    pass


@dataclasses.dataclass
class BaseArsItem(zf.Item):
    """
    Base class for all ``zelfred.Item`` subclasses in ``aws_resource_search`` project.
    """

    def open_url_or_print(self, ui: zf.UI, url: str):  # pragma: no cover
        """
        A helper user action method to open a URL in browser or print it to the terminal.

        Sometime user are in a remote shell and doesn't have default web browser.
        Then we print the url instead.

        .. note::

            This method will be used in the ``enter_handler``.
        """
        try:
            zf.open_url(url)
        # if remote shell may not have the ``open`` or ``start`` command
        # then FileNotFoundError will be raised
        except FileNotFoundError as e:
            if "start" in str(e) or "open" in str(e):
                print(
                    f"{ui.terminal.cyan}Your system doesn't support open url in browser, "
                    f"we print it here so you can copy manually.{ui.terminal.normal}"
                )
                print(url)
                ui.need_run_handler = False
                # since the print(url) will mess up the terminal UI,
                # we have to exit the app.
                # if we want to exit the session, based on the source code of
                # ``zf.UI.run_session``, we have to raise either
                # ``KeyboardInterrupt`` or ``zf.exc.EndOfInputError``.
                # either one would work, we intentionally leave both options here
                raise KeyboardInterrupt
                # raise zf.exc.EndOfInputError(selection=self)
            else:
                raise e

    def copy_or_print(self, ui: zf.UI, text: str):  # pragma: no cover
        """
        A helper user action method to copy a text to clipboard or print it to the terminal.

        Sometime user are in a remote shell and cannot use the clipboard.
        Then we print the value instead.

        .. note::

            This method will be used in the ``ctrl_a_handler`` or ``ctrl_w_handler``.
        """
        try:
            pyperclip.copy(text)
        except pyperclip.PyperclipException:
            print(
                f"{ui.terminal.cyan}Your system doesn't support copy to clipboard, "
                f"we print it here so you can copy manually.{ui.terminal.normal}"
            )
            print(text)
            ui.need_run_handler = False
            raise KeyboardInterrupt
            # raise zf.exc.EndOfInputError(selection=self)

    def open_file_or_print(self, ui: zf.UI, path: Path):  # pragma: no cover
        """
        A helper user action method to open a file in default application
        or print it to the terminal.

        Sometime user are in a remote shell and doesn't have an app to open a file.
        Then we print the url instead.

        .. note::

            This method will be used in the ``enter_handler``.
        """
        # try:
        zf.open_file(path)
        # except Exception as e:
        #     print(
        #         f"{ui.terminal.cyan}Your system doesn't support open a file, "
        #         f"we print it here, so you can see it.{ui.terminal.normal}"
        #     )
        #     print(path.read_text())
        #     ui.need_run_handler = False
        #     raise KeyboardInterrupt

    # by default, we don't quit the app when user press any action key
    # ``ui.wait_next_user_input()`` will wait for another user input and
    # skip repainting the UI in the next event loop
    def post_enter_handler(self, ui: zf.UI):  # pragma: no cover
        ui.wait_next_user_input()

    def post_ctrl_a_handler(self, ui: zf.UI):  # pragma: no cover
        ui.wait_next_user_input()

    def post_ctrl_w_handler(self, ui: zf.UI):  # pragma: no cover
        ui.wait_next_user_input()

    def post_ctrl_u_handler(self, ui: zf.UI):  # pragma: no cover
        ui.wait_next_user_input()

    def post_ctrl_p_handler(self, ui: zf.UI):  # pragma: no cover
        ui.wait_next_user_input()


T_ARS_ITEM = T.TypeVar("T_ARS_ITEM", bound=BaseArsItem)



# -*- coding: utf-8 -*-

import typing as T
import json

from .utils import guid, envs, rand_env

if T.TYPE_CHECKING:
    from .main import FakeAws


class CloudFormationMixin:
    cloudformation_stacks: T.List[str] = list()

    @classmethod
    def create_cloudformation_stack(cls: T.Type["FakeAws"]):
        tpl_data = {
            "AWSTemplateFormatVersion": "2010-09-09",
            "Parameters": {
                "BucketName": {
                    "Type": "String",
                }
            },
            "Resources": {
                "S3Bucket": {
                    "Type": "AWS::S3::Bucket",
                    "Properties": {
                        "BucketName": {"Ref": "BucketName"},
                    },
                },
            },
        }
        for ith in range(1, 1 + 10):
            env = rand_env()
            stack_name = f"{env}-{guid}-{ith}-cloudformation-stack"
            cls.bsm.cloudformation_client.create_stack(
                StackName=stack_name,
                TemplateBody=str(json.dumps(tpl_data)),
                Parameters=[
                    dict(
                        ParameterKey="BucketName",
                        ParameterValue=f"cft-managed-{ith}",
                    )
                ],
            )
            cls.cloudformation_stacks.append(stack_name)


# -*- coding: utf-8 -*-

from aws_resource_search.tests.fake_aws.api import FakeAws


class TestApi(FakeAws):
    @classmethod
    def setup_class_post_hook(cls):
        cls.setup_ars()

    def test(self):
        import aws_resource_search.api as aws_resource_search

        _ = aws_resource_search.format_shortcut
        _ = aws_resource_search.highlight_text
        _ = aws_resource_search.format_resource_type
        _ = aws_resource_search.format_key
        _ = aws_resource_search.format_value
        _ = aws_resource_search.format_key_value
        _ = aws_resource_search.ShortcutEnum
        _ = aws_resource_search.SUBTITLE
        _ = aws_resource_search.SHORT_SUBTITLE
        _ = aws_resource_search.SearcherEnum
        _ = aws_resource_search.T_RESULT_DATA
        _ = aws_resource_search.ResourceIterproxy
        _ = aws_resource_search.ResultPath
        _ = aws_resource_search.list_resources
        _ = aws_resource_search.extract_tags
        _ = aws_resource_search.BaseArsDocument
        _ = aws_resource_search.T_ARS_DOCUMENT
        _ = aws_resource_search.ResourceTypeDocument
        _ = aws_resource_search.get_utc_now
        _ = aws_resource_search.to_human_readable_elapsed
        _ = aws_resource_search.to_utc_dt
        _ = aws_resource_search.SIMPLE_DT_FMT
        _ = aws_resource_search.to_simple_dt_fmt
        _ = aws_resource_search.to_iso_dt_fmt
        _ = aws_resource_search.get_none_or_default
        _ = aws_resource_search.get_description
        _ = aws_resource_search.get_datetime
        _ = aws_resource_search.get_datetime_simple_fmt
        _ = aws_resource_search.get_datetime_iso_fmt
        _ = aws_resource_search.ResourceDocument
        _ = aws_resource_search.T_ARS_RESOURCE_DOCUMENT
        _ = aws_resource_search.BaseArsItem
        _ = aws_resource_search.T_ARS_ITEM
        _ = aws_resource_search.DetailItem
        _ = aws_resource_search.ExceptionItem
        _ = aws_resource_search.FileItem
        _ = aws_resource_search.InfoItem
        _ = aws_resource_search.UrlItem
        _ = aws_resource_search.AwsResourceTypeItem
        _ = aws_resource_search.AwsResourceItem
        _ = aws_resource_search.SetAwsProfileItem
        _ = aws_resource_search.ShowAwsInfoItem
        _ = aws_resource_search.preprocess_query
        _ = aws_resource_search.BaseSearcher
        _ = aws_resource_search.T_SEARCHER
        _ = aws_resource_search.config
        _ = aws_resource_search.ARS
        _ = aws_resource_search.search_aws_profile_handler
        _ = aws_resource_search.search_resource_type_handler
        _ = aws_resource_search.search_resource_handler
        _ = aws_resource_search.show_aws_info_handler
        _ = aws_resource_search.UI
        _ = aws_resource_search.handler


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(__file__, "aws_resource_search.api", preview=False)



# -*- coding: utf-8 -*-

"""
S3 related resources.

This module is an example of how to enable aws_resource_search to support new
resource types.
"""

import typing as T
import json
import dataclasses
from datetime import datetime

import aws_console_url.api as acu

from .. import res_lib as rl

if T.TYPE_CHECKING:
    from ..ars_def import ARS


# Declare a new class for the resource type "S3 bucket".
# It is a searchable object stored in the index, representing the
# metadata of an AWS resource.
# The class name should be "${ServiceName}${ResourceName}" in camel case.
# it should inherit from res_lib.BaseDocument.
@dataclasses.dataclass
class S3Bucket(rl.ResourceDocument):
    """
    S3 Bucket resource data model.
    """

    # declare the field names and types.
    # for most of AWS resource, it may have a "${resource_name}_arn" attribute.
    # most of AWS resource should store the ARN value as an attribute.
    # however, S3 bucket ARN can be calculated from the bucket name, so we
    # don't need to declare it here.
    @property
    def creation_date(self) -> T.Optional[datetime]:
        return self.raw_data.get("CreationDate")

    # it has to have a class method named "from_resource", it converts the
    # the s3 bucket data in the response of boto3 s3_client.list_buckets method
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/client/list_buckets.html
    # to the S3Bucket object.
    # note that some attributes may not be available in the response,
    # so we should program defensively to provide a default value
    # for example, we could do creation_date = resource.get("CreationDate", "NA")
    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id=resource["Name"],
            name=resource["Name"],
        )

    # it may have some additional property methods to provide more
    # human intuitive access to AWS resource attribute.
    @property
    def bucket_name(self) -> str:
        return self.name

    # it has to have a property named "title", it is the first line of the
    # search result item in the dropdown menu. you can format the title using
    # terminal format and color
    # (Query): s3-bucket: my-bucket
    # [x] bucket_name = my-bucket # <--- THIS IS title
    #       create_at = 2020-01-01T00:00:00+00:00, 🌐 Enter, 📋 Ctrl A, 🔗 Ctrl U, 👀 Ctrl P.
    @property
    def title(self) -> str:
        """
        Example: :cyan:`bucket_name` = :yellow:`my-bucket`
        """
        return rl.format_key_value("bucket_name", self.name)

    # it has to have a property named "subtitle", it is the second line of the
    # search result item in the dropdown menu. you can format the title using
    # terminal format and color
    # (Query): s3-bucket: my-bucket
    # [x] bucket_name = my-bucket
    #       create_at = 2020-01-01T00:00:00+00:00, 🌐 Enter, 📋 Ctrl A, 🔗 Ctrl U, 👀 Ctrl P. # <--- THIS IS subtitle
    @property
    def subtitle(self) -> str:
        """
        Example: :cyan:`create_at` = :yellow:`2021-07-06T15:04:40+00:00`,
        🌐 :magenta:`Enter`, 📋 :magenta:`Ctrl A`, 🔗 :magenta:`Ctrl U`, 👀 :magenta:`Ctrl P`.
        """
        return "{}, {}".format(
            rl.format_key_value("create_at", self.creation_date),
            self.short_subtitle,
        )

    # it has to have a property named "autocomplete", it defines the text to be
    # fill in the query input box when the user tap 'Tab' for autocomplete
    # usually, it is the identifier of the resource, in this case, it is the s3 bucket name
    @property
    def autocomplete(self) -> str:
        """
        Automatically enter the bucket name.
        """
        return self.name

    # most of the AWS resource support ARN,
    # most of the time, the boto3 API will return the ARN value in the response,
    # so you can store that in the class attribute and just reference it in this
    # property. if the boto3 API doesn't return the ARN value, you have to define
    # a method to calculate the ARN value.
    @property
    def arn(self) -> str:
        """
        Example: ``arn:aws:s3:::my-bucket``
        """
        return f"arn:aws:s3:::{self.name}"

    # most of the AWS resource support console url,
    # you have to define a method to calculate the console url.
    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.s3.get_console_url(bucket=self.name)

    # most of the AWS resource support list all resources of this type in AWS console,
    # you have to define a method to calculate the console url.
    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.s3.buckets

    # the get_details method returns a list of items to be displayed in the
    # resource details view when user tap 'Ctrl P'.
    # you may call some boto3 API to get more details about the resource.
    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        """
        Include s3 uri, s3 arn, bucket location and tags in details.
        """
        # the first code block is to initialize a detail_items list
        # using the class attribute
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        # get initial detail items
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars, arn_key="s3 arn")
        detail_items.append(from_detail("s3 uri", f"s3://{self.name}", url=url))

        # the second code block is to call boto3 API to get more details
        # we should wrap the code block with ``self.enrich_details(detail_items)``
        # context manager because we may not have the permission
        # the context manager will catch the exception and add a debug message
        # to tell the user that we don't have the permission to call the API
        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.s3_client.get_bucket_location(Bucket=self.name)
            location = res["LocationConstraint"]
            if not location:
                location = "us-east-1"
            detail_items.append(from_detail("location", location, url=url))

        # below, we call more API to get more information
        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.s3_client.get_bucket_versioning(Bucket=self.name)
            versioning = res.get("Status", "Not enabled yet")
            mfa_delete = res.get("MFADelete", "Not enabled yet")
            detail_items.extend([
                from_detail("versioning", versioning, url=url),
                from_detail("mfa_delete", mfa_delete, url=url),
            ])

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.s3_client.get_bucket_encryption(Bucket=self.name)
            rules = res.get("ServerSideEncryptionConfiguration", {}).get("Rules", [])
            if rules:
                rule = rules[0]
                sse_algorithm = rule.get("ApplyServerSideEncryptionByDefault", {}).get("SSEAlgorithm", "Unknown")
                kms_master_key_id = rule.get("ApplyServerSideEncryptionByDefault", {}).get("KMSMasterKeyID", "Unknown")
                bucket_key_enabled = rule.get("BucketKeyEnabled", "Unknown")
                detail_items.extend([
                    from_detail("sse_algorithm", sse_algorithm, url=url),
                    from_detail("kms_master_key_id", kms_master_key_id, url=url),
                    from_detail("bucket_key_enabled", bucket_key_enabled, url=url),
                ])

        # similar to the second code block
        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.s3_client.get_bucket_policy(Bucket=self.name)
            policy = res.get("Policy", "{}")
            detail_items.append(from_detail("bucket_policy", policy, self.one_line(policy), url=url))

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.s3_client.get_bucket_cors(Bucket=self.name)
            dct = {"CORSRules": res.get("CORSRules", [])}
            cors = json.dumps(dct)
            detail_items.append(from_detail("CORS", cors, self.one_line(cors), url=url))

        # the last code block is usually to get the tags of the resource
        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.s3_client.get_bucket_tagging(Bucket=self.name)
            tags = rl.extract_tags(res)
            detail_items.extend(rl.DetailItem.from_tags(tags, url))

        return detail_items
    # fmt: on


# create a res_lib.Searcher object. It defines the search behavior including
# how to get the data, how to index the data, and how to search the data.
class S3BucketSearcher(rl.BaseSearcher[S3Bucket]):
    pass


s3_bucket_searcher = S3BucketSearcher(
    # --- list resources
    # the s3 client argument in https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html
    service="s3",
    # the method name in https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/client/list_buckets.html
    method="list_buckets",
    # is the above method a paginator?
    is_paginator=False,
    # is there any other arguments to pass to the above method?
    # if it is a paginator, usually you need to define the pagination config
    # {"PaginationConfig": {"MaxItems": 9999, "PageSize": 1000}}
    default_boto_kwargs=None,
    # how to access the list of resource data in the API response
    # you can find it in the "Response Syntax" section of
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/client/list_buckets.html
    result_path=rl.ResultPath("Buckets"),
    # --- extract document
    # what is the corresponding document class for this AWS resource?
    doc_class=S3Bucket,
    # --- search
    # the resource type identifier in string, the naming convention is
    # ${service_name}-${resource_name}
    resource_type=rl.SearcherEnum.s3_bucket.value,
    # ``fields`` define how to index this document and how to search this document
    # the field list here has to match the attribute list in the document class
    # all field should be stored, so that we can recover the document object
    # from the search result
    fields=S3Bucket.get_dataset_fields(),
    # the list_buckets API result will be cached for 24 hours, so that we don't
    # need to rebuild the index every time when user search
    # user can use !~ query to force refresh the index
    # you can update this value to extend the cache expire time
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.s3_bucket.value),
    # this is only used for child resource, we will cover it in the sfn.py file
    more_cache_key=None,
)



# -*- coding: utf-8 -*-

"""
todo: add docstring
"""

from .base_document import BaseArsDocument
from .base_document import T_ARS_DOCUMENT
from .resource_type_document import ResourceTypeDocument
from .resource_document import get_utc_now
from .resource_document import to_human_readable_elapsed
from .resource_document import to_utc_dt
from .resource_document import SIMPLE_DT_FMT
from .resource_document import to_simple_dt_fmt
from .resource_document import to_iso_dt_fmt
from .resource_document import get_none_or_default
from .resource_document import get_description
from .resource_document import get_datetime
from .resource_document import get_datetime_simple_fmt
from .resource_document import get_datetime_iso_fmt
from .resource_document import ResourceDocument
from .resource_document import T_ARS_RESOURCE_DOCUMENT



# -*- coding: utf-8 -*-

"""
See :func:`select_resource_handler`.
"""

import typing as T

import botocore.exceptions
import zelfred.api as zf

from ..paths import path_aws_config, path_aws_credentials
from .. import res_lib as rl


if T.TYPE_CHECKING:  # pragma: no cover
    from ..ui_def import UI


def creating_index_items(resource_type: str) -> T.List[rl.InfoItem]:  # pragma: no cover
    """
    Print a message to tell user that we are creating index.

    This method is used when we need to create or recreate the index.
    """
    return [
        rl.InfoItem(
            title=f"Pulling data for {resource_type!r}, it may takes 1-60 seconds ...",
            subtitle="please wait, don't press any key",
            uid="pulling-data",
        )
    ]


T_DOC_TO_ITEM_FUNC = T.Callable[[rl.T_ARS_RESOURCE_DOCUMENT], rl.AwsResourceItem]


def search_resource_and_return_items(
    ui: "UI",
    searcher: rl.T_SEARCHER,
    query: str,
    boto_kwargs: T.Optional[dict] = None,
    refresh_data: bool = False,
    doc_to_item_func: T_DOC_TO_ITEM_FUNC = None,
    skip_ui: bool = False,
) -> T.List[T.Union[rl.AwsResourceItem, rl.InfoItem, rl.FileItem]]:
    """
    A wrapper of the :class:`~aws_resource_search.res_lib.Searcher`.
    It will return a list of UI items instead of a list of documents.

    :param searcher: corresponding resource type searcher object.
    :param query: example "my bucket", "my role".
    :param boto_kwargs: additional boto3 kwargs.
    :param refresh_data: do we need to refresh the data?
    :param doc_to_item_func: a callable function that convert document object
        to zelfred Item object. For those resources don't have mandatory
        arguments for boto3 API, it is very simple. But for those resources
        need a parent resource name as an argument for the boto3 API,
        we need special handling to convert the original doc to item.
        That's the purpose of this argument.
    :param skip_ui: if True, skip the UI related logic, just return the items.
        this argument is used for third party integration.
    """
    try:
        docs: T.List[rl.T_ARS_RESOURCE_DOCUMENT] = searcher.search(
            query=query,
            boto_kwargs=boto_kwargs,
            refresh_data=refresh_data,
        )
    except botocore.exceptions.ClientError as e:  # pragma: no cover
        return [
            rl.InfoItem(
                title=(
                    f"🔴 boto3 client error! "
                    f"check your default profile in ~/.aws/config and ~/.aws/credentials file."
                ),
                subtitle=repr(e),
                uid="boto3-client-error",
            ),
            rl.FileItem(
                title=f"Check ~/.aws/config",
                subtitle=f"{rl.ShortcutEnum.ENTER} to open file",
                arg=str(path_aws_config),
                uid="open-aws-config",
            ),
            rl.FileItem(
                title=f"Check ~/.aws/credentials",
                subtitle=f"{rl.ShortcutEnum.ENTER} to open file",
                arg=str(path_aws_credentials),
            ),
        ]

    # pprint(docs[:3]) # for DEBUG ONLY

    if doc_to_item_func is None:

        def doc_to_item_func(doc: rl.T_ARS_RESOURCE_DOCUMENT) -> rl.AwsResourceItem:
            return rl.AwsResourceItem.from_document(
                resource_type=searcher.resource_type,
                doc=doc,
            )

    items = [doc_to_item_func(doc=doc) for doc in docs]
    # pprint(items[:3]) # for DEBUG ONLY
    if len(items):
        return items
    else:
        # display helper text to tell user that we can't find any resource
        line = ui.line_editor.line
        if "@" in line:  # cannot find any sub resource
            autocomplete = line.split("@", 1)[0] + "@"
        else:  # cannot find resource
            autocomplete = line.split(":", 1)[0] + ": "
        return [
            rl.InfoItem(
                title=f"🔴 No {searcher.resource_type!r} found",
                subtitle=(
                    "Please try another query, "
                    "or type {} to refresh data, "
                    "or tap {} to clear existing query."
                ).format(
                    rl.highlight_text("!~"),
                    rl.ShortcutEnum.TAB,
                ),
                autocomplete=autocomplete,
            )
        ]


def search_resource(
    ui: "UI",
    resource_type: str,
    query: str,
    boto_kwargs: T.Optional[dict] = None,
    doc_to_item_func: T_DOC_TO_ITEM_FUNC = None,
    skip_ui: bool = False,
) -> T.List[T.Union[rl.AwsResourceItem, rl.InfoItem, rl.FileItem]]:
    """
    A wrapper around :func:`search_and_return_items`, it also handles the case
    that we have to create the index and the case that user wants to refresh data.

    .. note::

        this is the sub logic to handle AWS resource that doesn't have
        any mandatory boto kwargs. This is the MOST common case.

    :param ui: UI object.
    :param resource_type: the resource type name
    :param query: example "my bucket", "my role"
    :param doc_to_item_func: a callable function that convert document object
        to zelfred Item object.
    :param skip_ui: if True, skip the UI related logic, just return the items.
        this argument is used for third party integration.
    """
    zf.debugger.log(f"search_resource Query: {query}")
    final_query = rl.preprocess_query(query)
    searcher = ui.ars.get_searcher(resource_type)
    final_boto_kwargs = searcher._get_final_boto_kwargs(boto_kwargs=boto_kwargs)
    ds = searcher._get_ds(bsm=ui.ars.bsm, final_boto_kwargs=final_boto_kwargs)

    # display "creating index ..." message
    if ds.cache_key not in ds.cache:
        if skip_ui is False:  # pragma: no cover
            ui.run_handler(items=creating_index_items(resource_type))
            ui.repaint()
        return search_resource_and_return_items(
            ui=ui,
            searcher=searcher,
            query=final_query,
            boto_kwargs=boto_kwargs,
            doc_to_item_func=doc_to_item_func,
            skip_ui=skip_ui,
        )

    # manually refresh data
    if final_query.endswith("!~"):
        if skip_ui is False:  # pragma: no cover
            ui.run_handler(items=creating_index_items(resource_type))
            ui.repaint()
            ui.line_editor.press_backspace(n=2)
        return search_resource_and_return_items(
            ui=ui,
            searcher=searcher,
            query=rl.preprocess_query(final_query[:-2]),
            boto_kwargs=boto_kwargs,
            refresh_data=True,
            doc_to_item_func=doc_to_item_func,
            skip_ui=skip_ui,
        )

    # example: "ec2-inst: dev box"
    return search_resource_and_return_items(
        ui=ui,
        searcher=searcher,
        query=final_query,
        boto_kwargs=boto_kwargs,
        doc_to_item_func=doc_to_item_func,
        skip_ui=skip_ui,
    )


def search_partitioner(
    ui: "UI",
    resource_type: str,
    partitioner_resource_type: str,
    partitioner_query: str,
    skip_ui: bool = False,
) -> T.List[T.Union[rl.AwsResourceItem, rl.InfoItem, rl.FileItem]]:
    """
    Search partitioner resource.

    A wrapper of the :class:`~aws_resource_search.base_searcher.BaseSearcher`.
    It assumes that you have to search another resource before searching the
    real resource. For example, you have to search glue database before you can
    search glue table, since glue table boto3 API call requires the glue database.
    In this example, the glue_table is the searcher, and the glue_database is the
    partitioner_searcher. It returns a list of UI items instead of a list of documents.

    :param ui: UI object.
    :param resource_type: the resource type name
    :param partitioner_resource_type: the partitioner (parent) resource type name
    :param partitioner_query: the query to filter partitioner, example: "sfn state machine"
    :param skip_ui: if True, skip the UI related logic, just return the items.
        this argument is used for third party integration.
    """
    zf.debugger.log(f"search_partitioner Query: {partitioner_query!r}")

    def doc_to_item_func(doc: rl.T_ARS_RESOURCE_DOCUMENT) -> rl.AwsResourceItem:
        return rl.AwsResourceItem(
            uid=doc.uid,
            # title=f"{format_resource_type(partitioner_resource_type)}: {format_value(doc.autocomplete)}",
            title=f"{rl.format_resource_type(partitioner_resource_type)}: {doc.title}",
            subtitle=(
                f"Tap {rl.ShortcutEnum.TAB} "
                f"to search {rl.format_resource_type(resource_type)} "
                f"in this {rl.format_resource_type(partitioner_resource_type)}, "
                f"Tap {rl.ShortcutEnum.ENTER} to open {rl.format_resource_type(partitioner_resource_type)} url."
            ),
            autocomplete=f"{resource_type}: {doc.autocomplete}@",
            variables={
                "doc": doc,
                "resource_type": resource_type,
                "partitioner_resource_type": partitioner_resource_type,
            },
        )

    return search_resource(
        ui=ui,
        resource_type=partitioner_resource_type,
        query=partitioner_query,
        doc_to_item_func=doc_to_item_func,
        skip_ui=skip_ui,
    )


def search_child_resource(
    ui: "UI",
    resource_type: str,
    partitioner_resource_type: str,
    resource_query: str,
    partitioner_query: str,
    boto_kwargs: T.Dict[str, T.Any],
    skip_ui: bool = False,
) -> T.List[T.Union[rl.AwsResourceItem, rl.InfoItem, rl.FileItem]]:
    """
    Search child resource under the given partitioner.

    A wrapper of the :class:`~aws_resource_search.res_lib.Searcher`. But it will
    return a list of UI items instead of a list of documents.

    :param ui: UI object.
    :param resource_type: example: "glue-database-table"
    :param partitioner_resource_type: example: "glue-database"
    :param resource_query: example: "my table"
    :param partitioner_query: example: "my database"
    :param boto_kwargs: example: {"database_name": "my database"},
        for searching glue table under the given database
    :param skip_ui: if True, skip the UI related logic, just return the items.
        this argument is used for third party integration.
    """
    zf.debugger.log(f"search_child_resource Partitioner Query: {partitioner_query!r}")
    zf.debugger.log(f"search_child_resource Child Query: {resource_query!r}")

    def doc_to_item_func(doc: rl.T_ARS_RESOURCE_DOCUMENT) -> rl.AwsResourceItem:
        return rl.AwsResourceItem(
            uid=doc.uid,
            title=f"{rl.format_resource_type(resource_type)}: {doc.title}",
            subtitle=doc.subtitle,
            autocomplete=f"{resource_type}: {doc.autocomplete}",
            variables={
                "doc": doc,
                "resource_type": resource_type,
                "partitioner_resource_type": partitioner_resource_type,
            },
        )

    return search_resource(
        ui=ui,
        resource_type=resource_type,
        query=resource_query,
        boto_kwargs=boto_kwargs,
        doc_to_item_func=doc_to_item_func,
        skip_ui=skip_ui,
    )


def search_resource_under_partitioner(
    ui: "UI",
    resource_type: str,
    partitioner_resource_type: str,
    query: str,
    skip_ui: bool = False,
) -> T.List[T.Union[rl.AwsResourceItem, rl.InfoItem, rl.FileItem]]:
    """
    **IMPORTANT** this is the sub logic to handle AWS resource like glue table,
    glue job run, that requires the parent resource name in the arguments.

    :param resource_type: for example, ``"glue-table"``
    :param partitioner_resource_type: for example, ``"glue-database"``
    :param query: for example, if the full user query is ``"glue-table: my_database@my table"``,
        then this argument is ``"my_database@my table"``.
    :param skip_ui: if True, skip the UI related logic, just return the items.
        this argument is used for third party integration.
    """
    zf.debugger.log(f"search_resource_under_partitioner Query: {query}")
    q = zf.QueryParser(delimiter="@").parse(query)
    # --- search partitioner
    # for parent resource like glue job for glue job run, state machine for step function execution
    #
    # examples
    # - "  "
    # - "my database"

    # example: " "
    if len(q.trimmed_parts) == 0:
        return search_partitioner(
            resource_type=resource_type,
            partitioner_resource_type=partitioner_resource_type,
            partitioner_query="*",
            ui=ui,
            skip_ui=skip_ui,
        )
    # example: "my database "
    elif len(q.trimmed_parts) == 1 and len(q.parts) == 1:
        return search_partitioner(
            resource_type=resource_type,
            partitioner_resource_type=partitioner_resource_type,
            partitioner_query=q.trimmed_parts[0],
            ui=ui,
            skip_ui=skip_ui,
        )
    else:
        pass

    # --- search child resource
    # examples
    # - "my_database@"
    # - "my_database@my table"
    partitioner_query = q.trimmed_parts[0]
    boto_kwargs = ui.ars.get_partitioner_boto_kwargs(resource_type, partitioner_query)

    # example: "my_database@  "
    if len(q.trimmed_parts) == 1 and len(q.parts) > 1:
        return search_child_resource(
            ui=ui,
            resource_type=resource_type,
            partitioner_resource_type=partitioner_resource_type,
            resource_query="*",
            partitioner_query=partitioner_query,
            boto_kwargs=boto_kwargs,
            skip_ui=skip_ui,
        )
    # i.e. len(q.trimmed_parts) > 1
    # example: "my_database@my table"
    else:
        resource_query = query.split("@", 1)[1].strip()
        return search_child_resource(
            ui=ui,
            resource_type=resource_type,
            partitioner_resource_type=partitioner_resource_type,
            resource_query=resource_query,
            partitioner_query=partitioner_query,
            boto_kwargs=boto_kwargs,
            skip_ui=skip_ui,
        )


def search_resource_handler(
    ui: "UI",
    resource_type: str,
    query: str,
    skip_ui: bool = False,
) -> T.List[T.Union[rl.AwsResourceItem, rl.InfoItem, rl.FileItem]]:  # pragma: no cover
    """
    **IMPORTANT** This handle filter resource by query.

    :param resource_type: for example, ``"s3-bucket"``
    :param query: for example, if the full user query is ``"s3-bucket: my bucket"``,
        then this argument is ``"my bucket"``.
    :param skip_ui: if True, skip the UI related logic, just return the items.
        this argument is used for third party integration.
    """
    ui.render.prompt = f"(Query)"
    if ui.ars.has_partitioner(resource_type):
        return search_resource_under_partitioner(
            ui=ui,
            resource_type=resource_type,
            partitioner_resource_type=ui.ars.get_partitioner_resource_type(resource_type),
            query=query,
            skip_ui=skip_ui,
        )
    else:
        return search_resource(
            ui=ui,
            resource_type=resource_type,
            query=query,
            skip_ui=skip_ui,
        )



# -*- coding: utf-8 -*-

import typing as T
import copy
import dataclasses

from boto_session_manager import BotoSesManager
import sayt.api as sayt

from .paths import dir_index, dir_cache
from .utils import get_md5_hash
from .base_model import BaseModel
from .downloader import ResultPath, list_resources
from .documents.api import T_ARS_RESOURCE_DOCUMENT


SEP = "____"

T_MORE_CACHE_KEY = T.Callable[[sayt.T_DOCUMENT], T.List[str]]


def preprocess_query(query: T.Optional[str]) -> str:
    """
    Preprocess query, automatically add fuzzy search term if applicable.
    """
    delimiter = ".-_@+"
    if query:
        for char in delimiter:
            query = query.replace(char, " ")
        words = list()
        for word in query.split():
            if word.strip():
                word = word.strip()
                if len(word) == 1:
                    if word == "*":
                        words.append(word)
                else:
                    try:
                        if word[-2] != "~" and not word.endswith("!~"):
                            word = f"{word}~1"
                    except IndexError:
                        word = f"{word}~1"
                    words.append(word)
        if words:
            return " ".join(words)
        else:
            return "*"
    else:
        return "*"


@dataclasses.dataclass
class BaseSearcher(BaseModel, T.Generic[T_ARS_RESOURCE_DOCUMENT]):
    """
    todo: docstring

    :param service:
    :param method:
    :param is_paginator:
    :param default_boto_kwargs:
    :param result_path:
    :param doc_class:
    :param resource_type:
    :param fields:
    :param cache_expire:
    :param more_cache_key:
    :param bsm:
    """

    # list resources related
    service: str = dataclasses.field()
    method: str = dataclasses.field()
    is_paginator: bool = dataclasses.field()
    default_boto_kwargs: T.Optional[dict] = dataclasses.field()
    result_path: ResultPath = dataclasses.field()
    # extract document related
    doc_class: T.Type[T_ARS_RESOURCE_DOCUMENT] = dataclasses.field()
    # search related
    resource_type: str = dataclasses.field()
    fields: T.List[sayt.T_Field] = dataclasses.field()
    cache_expire: int = dataclasses.field()
    more_cache_key: T.Optional[T_MORE_CACHE_KEY] = dataclasses.field()
    # boto session
    bsm: T.Optional[BotoSesManager] = dataclasses.field(default=None)

    def _get_bsm(
        self,
        bsm: T.Optional[BotoSesManager] = None,
    ) -> BotoSesManager:
        """
        Preprocess the bsm arguments. Allow user to override the default bsm.
        """
        if bsm is None:
            final_bsm = self.bsm
        else:
            final_bsm = bsm
        if not isinstance(final_bsm, BotoSesManager):
            raise TypeError(f"bsm must be BotoSesManager, not {type(bsm)}")
        return final_bsm

    def _get_bsm_fingerprint(
        self,
        bsm: BotoSesManager,
    ) -> T.Tuple[str, T.Optional[str]]:
        """
        Get the logical unique fingerprint of the boto3 session. It will be
        used in the index name and cache key naming convention.
        """
        if str(bsm.profile_name) == "Sentinel('NOTHING')":
            account_or_profile = bsm.aws_account_id
        else:  # pragma: no cover
            account_or_profile = bsm.profile_name
        if bsm.aws_region is None:  # pragma: no cover
            region = "unknown-region"
        else:
            region = bsm.aws_region
        return account_or_profile, region

    def _get_final_boto_kwargs(self, boto_kwargs: T.Optional[dict] = None) -> dict:
        """
        Get the final boto3 api call kwargs by merging the default boto3 api
        and kwargs overrides
        """
        if self.default_boto_kwargs:
            final_boto_kwargs = copy.deepcopy(self.default_boto_kwargs)
        else:
            final_boto_kwargs = {}
        if boto_kwargs is not None:
            final_boto_kwargs.update(boto_kwargs)
        return final_boto_kwargs

    def _get_ds(
        self,
        bsm: BotoSesManager,
        final_boto_kwargs: dict,
    ) -> sayt.DataSet:
        """
        Get the corresponding ``sayt.DataSet`` object.
        """
        account_or_profile, region = self._get_bsm_fingerprint(bsm=bsm)
        if self.more_cache_key is None:
            index_name = SEP.join([account_or_profile, region, self.resource_type])
        else:
            index_name = SEP.join(
                [
                    account_or_profile,
                    region,
                    self.resource_type,
                    get_md5_hash(SEP.join(self.more_cache_key(final_boto_kwargs))),
                ]
            )
        cache_key = index_name
        cache_tag = index_name

        def downloader():
            resource_data_iter_proxy = list_resources(
                bsm=bsm,
                service=self.service,
                method=self.method,
                is_paginator=self.is_paginator,
                boto_kwargs=final_boto_kwargs,
                result_path=self.result_path,
            )
            for document in self.doc_class.from_many_resources(
                resources=resource_data_iter_proxy,
                bsm=bsm,
                boto_kwargs=final_boto_kwargs,
            ):
                doc_dict = document.to_dict()
                # print(doc_dict) # for DEBUG ONLY
                yield doc_dict

        return sayt.DataSet(
            dir_index=dir_index,
            index_name=index_name,
            fields=self.fields,
            dir_cache=dir_cache,
            cache_key=cache_key,
            cache_tag=cache_tag,
            cache_expire=self.cache_expire,
            downloader=downloader,
        )

    def search(
        self,
        query: str = "*",
        limit: int = 50,
        boto_kwargs: T.Optional[dict] = None,
        refresh_data: bool = False,
        simple_response: bool = True,
        verbose: bool = False,
        bsm: T.Optional[BotoSesManager] = None,
    ) -> T.Union[sayt.T_Result, T.List[T_ARS_RESOURCE_DOCUMENT]]:
        """
        Search the dataset.

        :param query: query string
        :param limit: the max number of results to return
        :param boto_kwargs: additional boto3 keyword arguments
        :param refresh_data: force to refresh the data
        :param simple_response: if True, then return a list of ``T_ARS_RESOURCE_DOCUMENT``
            objects, otherwise return the elasticsearch liked result.
        :param verbose: whether to print the log
        :param bsm: you can explicitly use a ``BotoSesManager`` object to override
            the default one you defined when creating the :class:`aws_resource_search.base_searcher.BaseSearcher`` object.
        """
        final_boto_kwargs = self._get_final_boto_kwargs(boto_kwargs=boto_kwargs)
        ds = self._get_ds(
            bsm=self._get_bsm(bsm),
            final_boto_kwargs=final_boto_kwargs,
        )
        final_query = preprocess_query(query)
        result = ds.search(
            query=final_query,
            limit=limit,
            simple_response=False,
            refresh_data=refresh_data,
            verbose=verbose,
        )
        if simple_response:
            return [self.doc_class.from_dict(dct["_source"]) for dct in result["hits"]]
        else:
            return result


T_SEARCHER = T.TypeVar("T_SEARCHER", bound=BaseSearcher)



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import sayt.api as sayt
import aws_arns.api as arns
import aws_console_url.api as acu

from .. import res_lib as rl

if T.TYPE_CHECKING:
    from ..ars_def import ARS


@dataclasses.dataclass
class EcsCluster(rl.ResourceDocument):
    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        ecs_cluster = arns.res.EcsCluster.from_arn(resource)
        return cls(
            raw_data=resource,
            id=ecs_cluster.cluster_name,
            name=ecs_cluster.cluster_name,
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("name", self.name)

    @property
    def subtitle(self) -> str:
        return "{}".format(
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.raw_data

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.ecs.get_cluster_services(name_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.ecs.clusters

    @property
    def cluster_name(self) -> str:
        return self.name

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.ecs_client.describe_clusters(
                clusters=[self.arn],
                include=["TAGS"],
            )
            dct = res.get("clusters", [None])[0]
            if dct:
                detail_items.extend(
                    [
                        from_detail("cluster_name", dct["clusterName"], url=url),
                        from_detail("cluster_arn", dct["clusterArn"], url=url),
                        from_detail("status", dct["status"], url=url),
                        from_detail("running_task_count", dct["runningTasksCount"], url=url),
                        from_detail("pending_task_count", dct["pendingTasksCount"], url=url),
                        from_detail("pending_task_count", dct["pendingTasksCount"], url=url),
                        from_detail("active_service_count", dct["activeServicesCount"], url=url),
                        from_detail("registered_container_instances_count", dct["registeredContainerInstancesCount"], url=url),
                    ]
                )
                tags = rl.extract_tags(res)
                detail_items.extend(rl.DetailItem.from_tags(tags, url=url))
        return detail_items
    # fmt: on


class EcsClusterSearcher(rl.BaseSearcher[EcsCluster]):
    pass


ecs_cluster_searcher = EcsClusterSearcher(
    # list resources
    service="ecs",
    method="list_clusters",
    is_paginator=True,
    default_boto_kwargs={"PaginationConfig": {"MaxItems": 9999, "PageSize": 100}},
    result_path=rl.ResultPath("clusterArns"),
    # extract document
    doc_class=EcsCluster,
    # search
    resource_type=rl.SearcherEnum.ecs_cluster.value,
    fields=EcsCluster.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.ecs_cluster.value),
    more_cache_key=None,
)


@dataclasses.dataclass
class EcsTaskRun(rl.ResourceDocument):
    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        task_run = arns.res.EcsTaskRun.from_arn(resource)
        return cls(
            raw_data=resource,
            id=task_run.short_id,
            name=task_run.short_id,
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("name", self.name)

    @property
    def subtitle(self) -> str:
        return "{}".format(
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return f"{self.cluster_name}@{self.name}"

    @property
    def arn(self) -> str:
        return self.raw_data

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.ecs.get_task_run_configuration(
            task_short_id_or_arn=self.arn, cluster_name=None
        )

    @property
    def cluster_name(self) -> str:
        task_run = arns.res.EcsTaskRun.from_arn(self.arn)
        return task_run.cluster_name

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.ecs_client.describe_tasks(
                cluster=self.cluster_name,
                tasks=[self.arn],
                include=["TAGS"],
            )
            task_dct = res.get("tasks", [None])[0]
            task_definition_arn = None
            if task_dct:
                task_definition_arn = task_dct["taskDefinitionArn"]
                detail_items.extend(
                    [
                        from_detail("cluster_name", self.cluster_name, url=ars.aws_console.ecs.get_cluster_tasks(name_or_arn=task_dct["clusterArn"])),
                        from_detail("task_definition_arn", task_definition_arn, url=ars.aws_console.ecs.get_task_definition_revision_containers(name_or_arn=task_definition_arn)),
                        from_detail("platform_family", task_dct.get("platformFamily", "NA"), url=url),
                        from_detail("platform_version", task_dct.get("platformVersion", "NA"), url=url),
                        from_detail("cpu", task_dct.get("cpu", "NA"), url=url),
                        from_detail("memory", task_dct.get("memory", "NA"), url=url),
                        from_detail("ephemeral_storage", task_dct.get("ephemeralStorage", {}).get("sizeInGiB", "NA"), url=url),
                        from_detail("started_at", task_dct.get("startedAt", "NA"), url=url),
                        from_detail("stopped_at", task_dct.get("stoppedAt", "NA"), url=url),
                        from_detail("started_by", task_dct.get("startedBy", "NA"), url=url),
                        from_detail("stop_code", task_dct.get("stopCode", "NA"), url=url),
                        from_detail("stopped_reason", task_dct.get("stoppedReason", "NA"), url=url),
                        from_detail("connectivity", task_dct.get("connectivity", "NA"), url=url),
                        from_detail("connectivity_at", task_dct.get("connectivityAt", "NA"), url=url),
                        from_detail("health_status", task_dct.get("healthStatus", "NA"), url=url),
                        from_detail("desired_status", task_dct.get("desiredStatus", "NA"), url=url),
                        from_detail("last_status", task_dct.get("lastStatus", "NA"), url=url),
                        from_detail("launch_type", task_dct.get("launchType", "NA"), url=url),
                    ]
                )

            if task_definition_arn is not None:
                res = ars.bsm.ecs_client.describe_task_definition(
                    taskDefinition=task_definition_arn,
                    include=["TAGS"],
                )
                task_def_dct = res.get("taskDefinition", None)
                if task_def_dct:
                    detail_items.extend(
                        [
                            from_detail("task_role_arn", task_def_dct["taskRoleArn"], url=ars.aws_console.iam.get_role(task_def_dct["taskRoleArn"])),
                            from_detail("task_execution_role_arn", task_def_dct["executionRoleArn"], url=ars.aws_console.iam.get_role(task_def_dct["executionRoleArn"])),
                        ]
                    )

            if task_dct:
                tags = rl.extract_tags(res)
                detail_items.extend(rl.DetailItem.from_tags(tags, url=url))

        return detail_items
    # fmt: on


class EcsTaskRunSearcher(rl.BaseSearcher[EcsTaskRun]):
    pass


ecs_task_run_searcher = EcsTaskRunSearcher(
    # list resources
    service="ecs",
    method="list_tasks",
    is_paginator=True,
    default_boto_kwargs={
        "PaginationConfig": {"MaxItems": 9999, "PageSize": 100},
    },
    result_path=rl.ResultPath("taskArns"),
    # extract document
    doc_class=EcsTaskRun,
    # search
    resource_type=rl.SearcherEnum.ecs_task_run.value,
    fields=EcsTaskRun.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.ecs_task_run.value),
    more_cache_key=None,
)


@dataclasses.dataclass
class EcsTaskDefinitionFamily(rl.ResourceDocument):
    # fmt: off
    v1_arn: str = dataclasses.field(metadata={"field": sayt.StoredField(name="v1_arn")})
    # fmt: on

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        ecs_task_def = arns.res.EcsTaskDefinition.new(
            aws_account_id=bsm.aws_account_id,
            aws_region=bsm.aws_region,
            task_name=resource,
            version=1,
        )
        return cls(
            raw_data=resource,
            id=ecs_task_def.task_name,
            name=ecs_task_def.task_name,
            v1_arn=ecs_task_def.to_arn(),
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("name", self.name)

    @property
    def subtitle(self) -> str:
        return "{}".format(
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.v1_arn

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.ecs.get_task_definition_revisions(name_or_arn=self.v1_arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.ecs.task_definitions

    @property
    def task_name(self) -> str:
        return self.name

    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.ecs_client.list_task_definitions(
                familyPrefix=self.name,
                sort="DESC",
                maxResults=100,
            )
            arn_lst = res.get("taskDefinitionArns", None)
            if arn_lst:
                detail_items.extend(
                    # fmt: off
                    [
                        from_detail("task_definition_arn", arn, url=ars.aws_console.ecs.get_task_definition_revision_containers(name_or_arn=arn))
                        for arn in arn_lst
                    ]
                    # fmt: on
                )
        return detail_items


class EcsTaskDefinitionFamilySearcher(rl.BaseSearcher[EcsTaskDefinitionFamily]):
    pass


ecs_task_definition_family_searcher = EcsTaskDefinitionFamilySearcher(
    # list resources
    service="ecs",
    method="list_task_definition_families",
    is_paginator=True,
    default_boto_kwargs={
        "status": "ACTIVE",
        "PaginationConfig": {"MaxItems": 9999, "PageSize": 100},
    },
    result_path=rl.ResultPath("families"),
    # extract document
    doc_class=EcsTaskDefinitionFamily,
    # search
    resource_type=rl.SearcherEnum.ecs_task_definition_family.value,
    fields=EcsTaskDefinitionFamily.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(
        rl.SearcherEnum.ecs_task_definition_family.value
    ),
    more_cache_key=None,
)



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import sayt.api as sayt
import aws_arns.api as arns
import aws_console_url.api as acu

from .. import res_lib as rl

if T.TYPE_CHECKING:
    from ..ars_def import ARS


@dataclasses.dataclass
class EcrRepository(rl.ResourceDocument):
    # fmt: off
    uri: str = dataclasses.field(metadata={"field": sayt.StoredField(name="uri")})
    # fmt: on

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id=resource["repositoryName"],
            name=resource["repositoryName"],
            uri=resource["repositoryUri"],
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("name", self.name)

    @property
    def subtitle(self) -> str:
        return "{}, {}".format(
            self.uri,
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.raw_data["repositoryArn"]

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.ecr.get_repo(name_or_arn_or_uri=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.ecr.repos

    @property
    def registry_id(self) -> str:
        return self.raw_data["registryId"]

    @property
    def repository_name(self) -> str:
        return self.raw_data["repositoryName"]

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        with rl.DetailItem.error_handling(detail_items):
            detail_items.extend(
                [
                    from_detail("registry_id", self.registry_id, url=url),
                    from_detail("repository_name", self.repository_name, url=url),
                    from_detail("repository_uri", self.raw_data["repositoryUri"], url=url),
                    from_detail("repository_arn", self.raw_data["repositoryArn"], url=url),
                    from_detail("create_at", self.raw_data["createdAt"], url=url),
                    from_detail("image_tag_mutability", self.raw_data["imageTagMutability"], url=url),
                ]
            )
            res = ars.bsm.ecr_client.get_repository_policy(
                registryId=self.registry_id,
                repositoryName=self.repository_name,
            )

            detail_items.extend([
                from_detail("repo_policy", self.one_line(res.get("policyText", "{}")), url=url),
            ])
            res = ars.bsm.ecr_client.list_tags_for_resource(resourceArn=self.arn)
            tags = rl.extract_tags(res)
            detail_items.extend(rl.DetailItem.from_tags(tags, url=url))
        return detail_items
    # fmt: on


class EcrRepositorySearcher(rl.BaseSearcher[EcrRepository]):
    pass


ecr_repository_searcher = EcrRepositorySearcher(
    # list resources
    service="ecr",
    method="describe_repositories",
    is_paginator=True,
    default_boto_kwargs={"PaginationConfig": {"MaxItems": 9999, "PageSize": 1000}},
    result_path=rl.ResultPath("repositories"),
    # extract document
    doc_class=EcrRepository,
    # search
    resource_type=rl.SearcherEnum.ecr_repository.value,
    fields=EcrRepository.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.ecr_repository.value),
    more_cache_key=None,
)


@dataclasses.dataclass
class EcrRepositoryImage(rl.ResourceDocument):
    # fmt: off
    repo_arn: str = dataclasses.field(metadata={"field": sayt.StoredField(name="repo_arn")})
    repo_uri: str = dataclasses.field(metadata={"field": sayt.StoredField(name="repo_uri")})
    # fmt: on

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        ecr_repo = arns.res.EcrRepository.new(
            aws_account_id=resource["registryId"],
            aws_region=bsm.aws_region,
            repo_name=resource["repositoryName"],
        )
        return cls(
            raw_data=resource,
            id=resource["imageDigest"].split(":")[-1],
            name=", ".join(resource.get("imageTags", [])),
            repo_arn=ecr_repo.to_arn(),
            repo_uri=ecr_repo.uri,
        )

    @property
    def title(self) -> str:
        return "{}, {}".format(
            rl.format_key_value("digest", self.id),
            rl.format_key_value("tags", self.name),
        )

    @property
    def subtitle(self) -> str:
        return "{}".format(
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return f"{self.repository_name}@{self.id}"

    @property
    def arn(self) -> str:
        return f"{self.repo_arn}:{self.id}"

    def get_console_url(self, console: acu.AWSConsole) -> str:
        # todo: add image url support to aws_console_url project
        origin_digest = self.raw_data["imageDigest"]
        return (
            f"https://{console.aws_region}.console.aws.amazon.com/ecr/repositories/private"
            f"/{self.registry_id}/{self.repository_name}/_"
            f"/image/{origin_digest}/details"
        )

    @property
    def digest(self) -> str:
        return self.id

    @property
    def registry_id(self) -> str:
        return self.raw_data["registryId"]

    @property
    def repository_name(self) -> str:
        return self.raw_data["repositoryName"]

    @property
    def image_tags(self) -> T.List[str]:
        return self.raw_data.get("imageTags", [])

    @property
    def uri(self) -> str:
        return f"{self.repo_uri}:{self.digest}"

    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        with rl.DetailItem.error_handling(detail_items):
            detail_items.extend(
                # fmt: off
                [
                    from_detail("registry_id", self.registry_id, url=url),
                    from_detail("repository_name", self.repository_name, url=url),
                    from_detail("repository_uri", self.repo_uri, url=url),
                    from_detail("repository_arn", self.repo_arn, url=url),
                    from_detail("image_pushed_at", self.raw_data["imagePushedAt"], url=url),
                    from_detail("last_recorded_pull_time", self.raw_data.get("lastRecordedPullTime", "NA"), url=url),
                    from_detail("image_scan_status", self.raw_data.get("imageScanStatus", {}).get("status", "NA"), url=url),
                ]
                # fmt: on
            )
            for image_tag in self.image_tags:
                detail_items.append(from_detail(f"image_tag", image_tag, url=url))
        return detail_items


class EcrRepositoryImageSearcher(rl.BaseSearcher[EcrRepositoryImage]):
    pass


ecr_repository_image_searcher = EcrRepositoryImageSearcher(
    # list resources
    service="ecr",
    method="describe_images",
    is_paginator=True,
    default_boto_kwargs={"PaginationConfig": {"MaxItems": 9999, "PageSize": 1000}},
    result_path=rl.ResultPath("imageDetails"),
    # extract document
    doc_class=EcrRepositoryImage,
    # search
    resource_type=rl.SearcherEnum.ecr_repository_image.value,
    fields=EcrRepositoryImage.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.ecr_repository_image.value),
    more_cache_key=lambda boto_kwargs: [boto_kwargs["repositoryName"]],
)



# -*- coding: utf-8 -*-

import typing as T

from .utils import guid, envs, rand_env

if T.TYPE_CHECKING:
    from .main import FakeAws


class GlueMixin:
    glue_databases: T.List[str] = list()
    glue_tables: T.List[str] = list()
    glue_jobs: T.List[str] = list()
    glue_job_runs: T.List[str] = list()
    glue_crawlers: T.List[str] = list()

    @classmethod
    def create_glue_database_table(cls: T.Type["FakeAws"]):
        for ith in range(6):
            env = rand_env()
            db_name = f"{env}-{guid}-{ith}-glue-database"
            cls.bsm.glue_client.create_database(
                DatabaseInput=dict(
                    Name=db_name,
                ),
            )
            cls.glue_databases.append(db_name)
            for jth in range(10):
                tb_name = f"{env}-{guid}-{jth}-glue-table"
                cls.bsm.glue_client.create_table(
                    DatabaseName=db_name,
                    TableInput=dict(
                        Name=tb_name,
                    ),
                )
                cls.glue_tables.append(f"{db_name}.{tb_name}")

    @classmethod
    def create_glue_job(cls: T.Type["FakeAws"]):
        for ith in range(10):
            env = rand_env()
            job_name = f"{env}-{guid}-{ith}-glue-job"
            cls.bsm.glue_client.create_job(
                Name=job_name,
                Role="arn:aws:iam::123456789012:role/AWSGlueServiceRoleDefault",
                Command={
                    "Name": "glueetl",
                },
            )
            cls.glue_jobs.append(job_name)

            for _ in range(3):
                res = cls.bsm.glue_client.start_job_run(JobName=job_name)
                job_run_id = res["JobRunId"]
                cls.glue_job_runs.append(job_run_id)

    @classmethod
    def create_glue_crawler(cls: T.Type["FakeAws"]):
        for ith in range(10):
            env = rand_env()
            crawler_name = f"{env}-{guid}-{ith}-glue-crawler"
            cls.bsm.glue_client.create_crawler(
                Name=crawler_name,
                Role="arn:aws:iam::123456789012:role/AWSGlueServiceRoleDefault",
                Targets=dict(S3Targets=[]),
            )
            cls.glue_crawlers.append(crawler_name)



# -*- coding: utf-8 -*-

"""
See :func:`search_aws_profile_handler`.
"""

import typing as T

import sayt.api as sayt
import awscli_mate.api as awscli_mate

from ..paths import dir_index, dir_cache, path_aws_config
from .. import res_lib as rl

if T.TYPE_CHECKING:  # pragma: no cover
    from ..ui_def import UI


def downloader() -> T.List[sayt.T_DOCUMENT]:
    pairs = awscli_mate.AWSCliConfig().extract_profile_and_region_pairs()
    return [
        {
            "profile": profile,
            "profile_ng": profile,
            "region_ng": region,
        }
        for profile, region in pairs
    ]


index_name = "aws-resource-search-aws-profiles"
aws_profile_dataset = sayt.DataSet(
    dir_index=dir_index,
    index_name=index_name,
    fields=[
        sayt.IdField(name="profile", stored=True),
        sayt.NgramWordsField(name="profile_ng", stored=True, minsize=2, maxsize=10),
        sayt.NgramWordsField(name="region_ng", stored=True, minsize=2, maxsize=10),
    ],
    dir_cache=dir_cache,
    cache_key=index_name,
    cache_tag=index_name,
    cache_expire=5 * 60,
    downloader=downloader,
)


def search_aws_profile_and_return_items(
    line_input: str,
    profile_query: str,
    refresh_data: bool = False,
) -> T.List[rl.SetAwsProfileItem]:
    docs = aws_profile_dataset.search(
        query=profile_query,
        limit=50,
        simple_response=True,
        refresh_data=refresh_data,
    )
    if len(docs):
        return rl.SetAwsProfileItem.from_many_profile_region_pairs(
            pairs=[(doc["profile_ng"], doc["region_ng"]) for doc in docs],
            autocomplete=line_input,
        )
    else:
        return [
            rl.FileItem.from_file(
                path_aws_config,
                title="❗ Cannot find any AWS profile in ~/.aws/config",
            )
        ]


def search_aws_profile_handler(
    ui: "UI",
    line_input: str,
    profile_query: str,
    skip_ui: bool = False,
) -> T.List[rl.SetAwsProfileItem]:  # pragma: no cover
    """
    Search AWS CLI profile in ``~/.aws/config`` file, order by similarity to the query.
    This function will be triggered when user type ``!@`` without leaving the app.

    :param line_input: the original query input by user before the ``!@``.
    :param profile_query: aws profile query, the text after ``!@``.

    :param skip_ui: if True, skip the UI related logic, just return the items.
        this argument is used for third party integration.
    """
    if skip_ui is False:  # pragma: no cover
        ui.render.prompt = f"(AWS Profile Query)"
    return search_aws_profile_and_return_items(
        line_input=line_input,
        profile_query=rl.preprocess_query(profile_query),
    )



# -*- coding: utf-8 -*-

"""
See :class:`AwsResourceItem`.
"""

import typing as T
import dataclasses

from ..compat import TypedDict
from ..terminal import remove_text_format

from .base_item import BaseArsItem

if T.TYPE_CHECKING:  # pragma: no cover
    import aws_console_url.api as acu

    from ..documents.resource_document import T_ARS_RESOURCE_DOCUMENT
    from ..ui_def import UI


@dataclasses.dataclass
class T_AWS_RESOURCE_ITEM_VARIABLES(TypedDict):
    """
    Type hint for the "variables" field in :class:`AwsResourceItem`.
    """

    doc: "T_ARS_RESOURCE_DOCUMENT"
    resource_type: str
    partitioner_resource_type: T.Optional[str]


@dataclasses.dataclass
class AwsResourceItem(BaseArsItem):
    """
    Represent an item in the resource search result.

    :param variables: in AwsResourceItem, the variable is a dictionary including
        the original document object (not dict).
    """

    variables: T_AWS_RESOURCE_ITEM_VARIABLES = dataclasses.field(default_factory=dict)

    @classmethod
    def from_document(
        cls,
        resource_type: str,
        doc: "T_ARS_RESOURCE_DOCUMENT",
    ):
        """
        A factory method that convert a dictionary view of a
        :class:`~aws_resource_search.documents.resource_document.ResourceDocument`
        object to an :class:`AwsResourceItem`.

        For example: if the doc is::

            >>> doc = dataclasses.asdict(
            ...     S3Bucket(
            ...         title="my-bucket",
            ...         subtitle="bucket creation date",
            ...         uid="my-bucket",
            ...         autocomplete="my-bucket",
            ...     )
            ... )

        Then the item will be::

            >>> dataclasses.asdict(AwsResourceItem.from_document(doc))
            {
                "title": "my-bucket",
                "subtitle": "bucket creation date",
                "uid": "s3-bucket",
                "autocomplete": "s3-bucket: my-bucket",
                "variables": {"doc": S3Bucket(...)},
            }

        We also have an array version :meth:`AwsResourceItem.from_many_document`.
        """
        return cls(
            title=doc.title,
            subtitle=doc.subtitle,
            uid=doc.uid,
            autocomplete=f"{resource_type}: {doc.autocomplete}",
            variables={
                "doc": doc,
                "resource_type": resource_type,
                "partitioner_resource_type": None,
            },
        )

    @classmethod
    def from_many_document(
        cls,
        resource_type: str,
        docs: T.Iterable["T_ARS_RESOURCE_DOCUMENT"],
    ):
        """
        An array version of :meth:`AwsResourceItem.from_document`.
        """
        return [cls.from_document(resource_type, doc) for doc in docs]

    def get_id(self) -> str:
        """
        Get the resource id.
        """
        return self.variables["doc"].id

    def get_name(self) -> str:
        """
        Get the resource name.
        """
        return self.variables["doc"].name

    def get_console_url(self, console: "acu.AWSConsole") -> str:
        """
        Get AWS console url of the resource.
        """
        doc: "T_ARS_RESOURCE_DOCUMENT" = self.variables["doc"]
        return doc.get_console_url(console=console)

    def get_arn(self) -> str:
        """
        Get ARN of the resource.
        """
        return self.variables["doc"].arn

    def enter_handler(self, ui: "UI"):  # pragma: no cover
        """
        Default behavior:

        Open AWS console url in browser.
        """
        self.open_url_or_print(ui, self.get_console_url(console=ui.ars.aws_console))

    def ctrl_a_handler(self, ui: "UI"):  # pragma: no cover
        """
        Default behavior:

        Copy ARN to clipboard.
        """
        self.copy_or_print(ui, self.get_arn())

    def ctrl_u_handler(self, ui: "UI"):  # pragma: no cover
        """
        Default behavior:

        Copy AWS console url to clipboard.
        """
        self.copy_or_print(ui, self.get_console_url(console=ui.ars.aws_console))

    def ctrl_p_handler(self, ui: "UI"):  # pragma: no cover
        """
        View details in a sub session. You can tap 'F1' to exit the sub session.
        """
        doc: "T_ARS_RESOURCE_DOCUMENT" = self.variables["doc"]
        items = doc.get_details(ars=ui.ars)
        ui.run_handler(items=items)

        # enter the main event loop of the sub query
        # user can tap 'F1' to exit the sub query session,
        # and go back to the folder selection session.
        def handler(query: str, ui: "UI"):
            """
            A partial function that using the given folder.
            """
            ui.render.prompt = "(Detail)"
            return items

        ui.run_sub_session(
            handler=handler,
            initial_query=(
                f"{remove_text_format(self.title)}, " f"press F1 to go back."
            ),
        )



# -*- coding: utf-8 -*-

from aws_resource_search.handlers.search_aws_profile_handler import (
    search_aws_profile_and_return_items,
)
from aws_resource_search.tests.mock_aws_cli import test_home_aws_folder


def test_search_aws_profile_handler():
    with test_home_aws_folder.temp():
        items = search_aws_profile_and_return_items(
            line_input="s3-bucket: my-bucket",
            profile_query="*",
            refresh_data=True,
        )
        assert len(items) == 3

        items = search_aws_profile_and_return_items(
            line_input="s3-bucket: my-bucket",
            profile_query="*",
        )
        assert len(items) == 3

        items = search_aws_profile_and_return_items(
            line_input="s3-bucket: my-bucket",
            profile_query="invalidprofile",
        )
        assert len(items) == 1


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(
        __file__,
        "aws_resource_search.handlers.search_aws_profile_handler",
        preview=False,
    )



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import aws_arns.api as arns

import aws_console_url.api as acu

from .. import res_lib as rl

if T.TYPE_CHECKING:
    from ..ars_def import ARS


lambda_function_state_icon = {
    "Pending": "🟡",
    "Active": "🟢",
    "Inactive": "⚫",
    "Failed": "🔴",
}


@dataclasses.dataclass
class LambdaFunction(rl.ResourceDocument):
    @property
    def description(self) -> str:
        return rl.get_description(self.raw_data, "Description")

    @property
    def runtime(self) -> str:
        return self.raw_data.get("Runtime", "NA")

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id=resource["FunctionName"],
            name=resource["FunctionName"],
        )

    @property
    def func_name(self) -> str:
        return self.name

    @property
    def title(self) -> str:
        return rl.format_key_value("function_name", self.name)

    @property
    def subtitle(self) -> str:
        return "{}, {}, {}".format(
            rl.format_key_value("runtime", self.runtime),
            rl.format_key_value("description", self.description),
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.raw_data["FunctionArn"]

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.awslambda.get_function(name_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.awslambda.functions

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.lambda_client.get_function(FunctionName=self.name)
            func_config = res["Configuration"]
            description = rl.get_description(func_config, "Description")
            role_arn = func_config["Role"]
            runtime = func_config.get("Runtime", "NA")
            timeout = func_config["Timeout"]
            memory_size = func_config["MemorySize"]
            handler = func_config["Handler"]
            state = func_config.get("State", "NA")
            last_modified = func_config["LastModified"]
            architectures = func_config.get("Architectures", [])

            state_icon = lambda_function_state_icon[state]
            detail_items.extend([
                from_detail("description", description, url=url),
                from_detail("🧢 role_arn", role_arn, url=ars.aws_console.iam.get_role(role_arn)),
                from_detail("runtime", runtime, url=url),
                from_detail("timeout", timeout, url=url),
                from_detail("memory_size", memory_size, url=url),
                from_detail("handler", handler, url=url),
                from_detail("state", f"{state_icon} {state}", url=url),
                from_detail("last_modified", last_modified, url=url),
                from_detail("architectures", architectures, url=url),
            ])

            detail_items.extend([
                from_detail(
                    "layer",
                    dct["Arn"],
                    url=ars.aws_console.awslambda.get_layer(name_or_arn=dct["Arn"]),
                )
                for dct in func_config.get("Layers", [])
            ])
            env_vars = func_config.get("Environment", {}).get("Variables", {})
            detail_items.extend(rl.DetailItem.from_env_vars(env_vars, url))

            tags = rl.extract_tags(res)
            detail_items.extend(rl.DetailItem.from_tags(tags, url))

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.lambda_client.list_event_source_mappings(
                FunctionName=self.arn,
            )
            mappings = res.get("EventSourceMappings", [])
            if len(mappings):
                for mapping in mappings:
                    event_source_uuid = mapping["UUID"]
                    event_source_arn = mapping["EventSourceArn"]
                    state = mapping["State"]
                    detail_items.append(
                        rl.DetailItem.new(
                            title="mapping: {}, {}".format(
                                rl.format_key_value("event_source_arn", event_source_arn),
                                rl.format_key_value("state", state),
                            ),
                            subtitle=f"🌐 {rl.ShortcutEnum.ENTER} to open event source url, 📋 {rl.ShortcutEnum.CTRL_A} to copy event sourec arn.",
                            uid=event_source_uuid,
                            copy=event_source_arn,
                            url=arns.Arn.from_arn(event_source_arn).to_console_url(),
                        )
                    )
            else:
                detail_items.append(
                    rl.DetailItem.new(
                        title=f"🔴 No mapping found",
                        subtitle=f"{rl.ShortcutEnum.ENTER} to confirm in AWS console",
                        uid=f"no mapping found",
                        url=url,
                    )
                )

        return detail_items
    # fmt: on


class LambdaFunctionSearcher(rl.BaseSearcher[LambdaFunction]):
    pass


lambda_function_searcher = LambdaFunctionSearcher(
    # list resources
    service="lambda",
    method="list_functions",
    is_paginator=True,
    default_boto_kwargs={"PaginationConfig": {"MaxItems": 9999, "PageSize": 1000}},
    result_path=rl.ResultPath("Functions"),
    # extract document
    doc_class=LambdaFunction,
    # search
    resource_type=rl.SearcherEnum.lambda_function.value,
    fields=LambdaFunction.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.lambda_function.value),
    more_cache_key=None,
)


@dataclasses.dataclass
class LambdaFunctionAlias(rl.ResourceDocument):
    @property
    def alias(self) -> str:
        return self.name

    @property
    def function_name(self) -> str:
        return self.arn.split(":")[-2]

    @property
    def description(self) -> str:
        return rl.get_description(self.raw_data, "Description")

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        alias_arn = resource["AliasArn"]
        lbd_func = arns.res.LambdaFunction.from_arn(alias_arn)
        return cls(
            raw_data=resource,
            id=lbd_func.alias,
            name=lbd_func.alias,
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("alias", self.name)

    @property
    def subtitle(self) -> str:
        return "{}, {}".format(
            self.description,
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return f"{self.function_name}@{self.name}"

    @property
    def arn(self) -> str:
        return self.raw_data["AliasArn"]

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.awslambda.get_function_alias(name_or_arn=self.arn)

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.lambda_client.get_function_configuration(
                FunctionName=self.function_name,
                Qualifier=self.name,
            )
            description = res.get("Description", "NA")
            role_arn = res["Role"]
            runtime = res["Runtime"]
            timeout = res["Timeout"]
            memory_size = res["MemorySize"]
            handler = res["Handler"]
            state = res.get("State", "NA")
            last_modified = res["LastModified"]
            architectures = res.get("Architectures", [])

            state_icon = lambda_function_state_icon[state]
            detail_items.extend(
                [
                    from_detail("description", description, url=url),
                    from_detail("🧢 role_arn", role_arn, url=ars.aws_console.iam.get_role(role_arn)),
                    from_detail("runtime", runtime, url=url),
                    from_detail("timeout", timeout, url=url),
                    from_detail("memory_size", memory_size, url=url),
                    from_detail("handler", handler, url=url),
                    from_detail("state", f"{state_icon} {state}", url=url),
                    from_detail("last_modified", last_modified, url=url),
                    from_detail("architectures", architectures, url=url),
                ]
            )

            env_vars = res.get("Environment", {}).get("Variables", {})
            detail_items.extend(rl.DetailItem.from_env_vars(env_vars, url))

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.lambda_client.list_event_source_mappings(
                FunctionName=self.arn,
            )
            mappings = res.get("EventSourceMappings", [])
            if len(mappings):
                for mapping in mappings:
                    event_source_uuid = mapping["UUID"]
                    event_source_arn = mapping["EventSourceArn"]
                    state = mapping["State"]
                    detail_items.append(
                        rl.DetailItem.new(
                            title="mapping: {}, {}".format(
                                rl.format_key_value("event_source_arn", event_source_arn),
                                rl.format_key_value("state", state),
                            ),
                            subtitle=f"🌐 {rl.ShortcutEnum.ENTER} to open event source url, 📋 {rl.ShortcutEnum.CTRL_A} to copy event sourec arn.",
                            uid=event_source_uuid,
                            copy=event_source_arn,
                            url =arns.Arn.from_arn(event_source_arn).to_console_url(),
                        )
                    )
            else:
                detail_items.append(
                    rl.DetailItem.new(
                        title=f"🔴 No mapping found",
                        subtitle=f"{rl.ShortcutEnum.ENTER} to confirm in AWS console",
                        uid=f"no mapping found",
                        url=url,
                    )
                )

        return detail_items
    # fmt: on


class LambdaFunctionAliasSearcher(rl.BaseSearcher[LambdaFunctionAlias]):
    pass


lambda_function_alias_searcher = LambdaFunctionAliasSearcher(
    # list resources
    service="lambda",
    method="list_aliases",
    is_paginator=True,
    default_boto_kwargs={"PaginationConfig": {"MaxItems": 9999, "PageSize": 1000}},
    result_path=rl.ResultPath("Aliases"),
    # extract document
    doc_class=LambdaFunctionAlias,
    # search
    resource_type=rl.SearcherEnum.lambda_function_alias.value,
    fields=LambdaFunctionAlias.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(
        rl.SearcherEnum.lambda_function_alias.value
    ),
    more_cache_key=lambda boto_kwargs: [boto_kwargs["FunctionName"]],
)


@dataclasses.dataclass
class LambdaLayer(rl.ResourceDocument):
    @property
    def description(self) -> str:
        return rl.get_description(
            self.raw_data,
            "LatestMatchingVersion.Description",
        )

    @property
    def compatible_runtimes(self) -> T.List[str]:
        return rl.get_none_or_default(
            self.raw_data,
            "LatestMatchingVersion.CompatibleRuntimes",
            [],
        )

    @property
    def compatible_architectures(self) -> T.List[str]:
        return rl.get_none_or_default(
            self.raw_data,
            "LatestMatchingVersion.CompatibleArchitectures",
            [],
        )

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id=resource["LayerName"],
            name=resource["LayerName"],
        )

    @property
    def title(self) -> str:
        return self.name

    @property
    def subtitle(self) -> str:
        return "{}, {}, {}, {}".format(
            rl.format_key_value("runtime", self.compatible_runtimes),
            rl.format_key_value("arch", self.compatible_architectures),
            rl.format_key_value("description", self.description),
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.raw_data["LayerArn"]

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.awslambda.get_layer(name_or_arn=self.arn + ":1")

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.awslambda.layers


class LambdaLayerSearcher(rl.BaseSearcher[LambdaLayer]):
    pass


lambda_layer_searcher = LambdaLayerSearcher(
    # list resources
    service="lambda",
    method="list_layers",
    is_paginator=True,
    default_boto_kwargs={"PaginationConfig": {"MaxItems": 1000, "PageSize": 50}},
    result_path=rl.ResultPath("Layers"),
    # extract document
    doc_class=LambdaLayer,
    # search
    resource_type=rl.SearcherEnum.lambda_layer.value,
    fields=LambdaLayer.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.lambda_layer.value),
    more_cache_key=None,
)



# -*- coding: utf-8 -*-

from rich import print as rprint

from aws_resource_search.searcher_enum import SearcherEnum
from aws_resource_search.logger import logger
from aws_resource_search.tests.fake_aws.api import FakeAws, guid, envs, rand_env


class TestARSInit(FakeAws):
    @classmethod
    def setup_class_post_hook(cls):
        cls.setup_ars()
        cls.create_all()

    def _test_ars(self):
        ars = self.ars

        assert ars.is_valid_resource_type(SearcherEnum.s3_bucket.value) is True
        assert ars.is_valid_resource_type("not-available") is False

    def _test_all(self):
        """
        这个方法用来测试 list resource 的时候无需任何额外参数的情况, 也就是简单情况.
        """
        resource_type_list = [
            SearcherEnum.cloudformation_stack,
            SearcherEnum.dynamodb_table,
            SearcherEnum.ec2_instance,
            SearcherEnum.ec2_vpc,
            SearcherEnum.ec2_subnet,
            SearcherEnum.ec2_security_group,
            SearcherEnum.glue_crawler,
            SearcherEnum.glue_database,
            SearcherEnum.glue_job,
            # SearcherEnum.glue_job_run, # we need to test it in another way
            # SearcherEnum.glue_table, # we need to test it in another way
            SearcherEnum.iam_group,
            SearcherEnum.iam_policy,
            SearcherEnum.iam_role,
            SearcherEnum.iam_user,
            SearcherEnum.kms_key_alias,
            SearcherEnum.lambda_function,
            SearcherEnum.lambda_layer,
            SearcherEnum.s3_bucket,
            SearcherEnum.secretsmanager_secret,
            # SearcherEnum.sfn_state_machine_execution,
            SearcherEnum.sfn_state_machine,
            SearcherEnum.sns_topic,
            SearcherEnum.sqs_queue,
            SearcherEnum.ssm_parameter,
        ]

        ars = self.ars

        for resource_type in resource_type_list:
            logger.ruler(f"start {resource_type!r}")
            searcher = ars.get_searcher(resource_type=resource_type)
            result = searcher.search(
                query=guid,
                refresh_data=True,
                simple_response=True,
                verbose=False,
            )
            # logger.info(result)
            for doc in result:
                logger.info(
                    f"id = {doc.id!r} "
                    f"| name = {doc.name!r} "
                    f"| url = {doc.get_console_url(ars.aws_console)}"
                )
                assert guid in doc.name
            logger.ruler(f"end {resource_type!r}")

    def test(self):
        print("")
        with logger.disabled(
            disable=True,  # no log,
            # disable=False,  # show log
        ):
            self._test_ars()
            self._test_all()


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(__file__, "aws_resource_search.ars_def", preview=False)



# -*- coding: utf-8 -*-

import typing as T
import random

from .utils import guid, envs, rand_env

if T.TYPE_CHECKING:
    from .main import FakeAws


class DynamodbMixin:
    dynamodb_tables: T.List[str] = list()

    @classmethod
    def create_dynamodb_tables(cls: T.Type["FakeAws"]):
        for ith in range(1, 1 + 10):
            env = rand_env()
            table = f"{env}-{guid}-{ith}-dynamodb-table"
            kwargs = dict(
                TableName=table,
                AttributeDefinitions=[
                    dict(
                        AttributeName="pk",
                        AttributeType="S",
                    ),
                ],
                KeySchema=[
                    dict(
                        AttributeName="pk",
                        KeyType="HASH",
                    )
                ],
                BillingMode="PAY_PER_REQUEST",
            )
            # 70% table only has hash key, 30% table also has sort key
            if random.randint(1, 100) <= 30:
                kwargs["AttributeDefinitions"].append(
                    dict(
                        AttributeName="sk",
                        AttributeType="S",
                    )
                )
                kwargs["KeySchema"].append(
                    dict(
                        AttributeName="sk",
                        KeyType="RANGE",
                    )
                )
            cls.bsm.dynamodb_client.create_table(**kwargs)
            cls.dynamodb_tables.append(table)


# -*- coding: utf-8 -*-

"""
See :class:`UrlItem`.
"""

import typing as T
import dataclasses

import zelfred.api as zf

from ..terminal import ShortcutEnum
from ..compat import TypedDict
from .base_item import BaseArsItem


class T_URL_ITEM_VARIABLES(TypedDict):
    url: str


@dataclasses.dataclass
class UrlItem(BaseArsItem):
    """
    Represent an url item in the dropdown menu.
    """

    variables: T_URL_ITEM_VARIABLES = dataclasses.field(default_factory=dict)

    def enter_handler(self, ui: "zf.UI"):  # pragma: no cover
        """
        Behavior:

        - If we have an url, open it in the default browser.
        - If not, nothing happens (most likely NOT)
        """
        if self.variables.get("url") is not None:
            self.open_url_or_print(ui, self.variables["url"])
        else:
            pass

    def ctrl_a_handler(self, ui: "zf.UI"):  # pragma: no cover
        """
        Behavior:

        - Copy the url to clipboard
        """
        self.ctrl_u_handler(ui)

    def ctrl_u_handler(self, ui: "zf.UI"):  # pragma: no cover
        """
        Behavior:

        - Copy the url to clipboard
        """
        if self.variables.get("url") is not None:
            self.copy_or_print(ui, str(self.variables["url"]))
        else:
            pass

    @classmethod
    def from_url(
        cls,
        url: str,
        title: T.Optional[str] = None,
        subtitle: T.Optional[str] = None,
        uid: T.Optional[str] = None,
        arg: T.Optional[str] = None,
        autocomplete: T.Optional[str] = None,
    ):
        """
        Create one :class:`FileItem` from an ``pathlib.Path`` object.
        """
        if title is None:
            title = f"🌐 {url}"
        if subtitle is None:
            subtitle = f"🌐 {ShortcutEnum.ENTER} to  {ShortcutEnum.CTRL_A} or {ShortcutEnum.CTRL_U} to copy url"
        kwargs = dict(
            title=title,
            subtitle=subtitle,
            arg=arg,
            autocomplete=autocomplete,
            variables={"url": url},
        )
        if uid:
            kwargs["uid"] = uid
        return cls(**kwargs)



# -*- coding: utf-8 -*-

"""
Usage::

    from aws_resource_search.tests.fake_aws.api import FakeAws, guid, envs, rand_env
"""

from .main import FakeAws
from .utils import guid
from .utils import envs
from .utils import rand_env


# -*- coding: utf-8 -*-

"""
See :class:`ResourceTypeDocument`.
"""

import typing as T
import dataclasses

import sayt.api as sayt

from .base_document import BaseArsDocument


@dataclasses.dataclass
class ResourceTypeDocument(BaseArsDocument):
    """
    This is the AWS resource type document class. A resource type is a type of
    AWS resource, like s3-bucket, ec2-instance, etc.

    :param id: the maintainer defined unique identifier of the resource type,
        using hyphen-case.
    :param name: same as the ``id`` field, but using ``NgramWordsField`` for searching
    :param desc: the maintainer defined description for the specific resource type.
    :param ngram: ngram search string for the resource type.
    """

    id: str = dataclasses.field()
    name: str = dataclasses.field()
    desc: str = dataclasses.field()
    ngram: str = dataclasses.field()

    @classmethod
    def get_dataset_fields(cls) -> T.List[sayt.T_Field]:
        return [
            sayt.IdField(name="id", stored=True),
            sayt.NgramWordsField(name="name", stored=True, minsize=2, maxsize=20),
            sayt.StoredField(name="desc"),
            sayt.NgramWordsField(name="ngram", stored=True, minsize=2, maxsize=20),
        ]



# -*- coding: utf-8 -*-

from aws_resource_search.items.set_aws_profile_item import SetAwsProfileItem


class TestSetAwsProfileItem:
    def test_from_profile_region(self):
        items = SetAwsProfileItem.from_many_profile_region_pairs(
            [("my_profile", "us-east-1")],
            autocomplete="",
        )
        # print(items)


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(
        __file__,
        "aws_resource_search.items.set_aws_profile_item",
        preview=False,
    )



# -*- coding: utf-8 -*-

from aws_resource_search.documents.resource_type_document import ResourceTypeDocument
from aws_resource_search.items.aws_resource_type_item import AwsResourceTypeItem
from aws_resource_search.tests.fake_aws.api import FakeAws


class TestAwsResourceTypeItem(FakeAws):
    @classmethod
    def setup_class_post_hook(cls):
        cls.setup_ars()

    def test_from_many_document(self):
        item = AwsResourceTypeItem.from_many_document(
            docs=[
                ResourceTypeDocument(
                    id="s3-bucket",
                    name="s3-bucket",
                    desc="s3-bucket",
                    ngram="s3-bucket",
                ),
            ],
            ars=self.ars,
        )[0]
        assert isinstance(item, AwsResourceTypeItem)


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(
        __file__,
        "aws_resource_search.items.aws_resource_type_item",
        preview=False,
    )



# -*- coding: utf-8 -*-

from aws_resource_search.handlers.search_resource_handler import (
    search_resource,
    search_resource_under_partitioner,
)
from aws_resource_search.tests.fake_aws.utils import guid
from aws_resource_search.tests.fake_aws.main import FakeAws


class TestSearchResourceHandler(FakeAws):
    @classmethod
    def setup_class_post_hook(cls):
        cls.setup_ars()
        cls.setup_ui()
        cls.create_s3_bucket()
        cls.create_state_machines()

    def test_search_resource(self):
        items = search_resource(
            ui=self.ui,
            resource_type="s3-bucket",
            query=guid,
            skip_ui=True,
        )
        assert len(items) > 1
        for item in items:
            assert guid in item.get_name()

        items = search_resource(
            ui=self.ui,
            resource_type="s3-bucket",
            query="invalidresource",
            skip_ui=True,
        )
        assert len(items) == 1

        items = search_resource(
            ui=self.ui,
            resource_type="s3-bucket",
            query=f"{guid}!~",
            skip_ui=True,
        )
        for item in items:
            assert guid in item.get_name()

    def test_search_resource_under_partitioner(self):
        sm_items = search_resource(
            ui=self.ui,
            resource_type="sfn-state-machine",
            query=guid,
            skip_ui=True,
        )
        for sm_item in sm_items:
            items = search_resource_under_partitioner(
                ui=self.ui,
                resource_type="sfn-state-machine-execution",
                partitioner_resource_type=f"sfn-state-machine",
                query=f"{sm_item.get_name()}@{guid}",
                skip_ui=True,
            )
            # this won't work because moto doesn't handle create_state_machine properly
            # the list_executions API doesn't work.
            # for item in items:
            #     assert guid in item.get_name()


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(
        __file__,
        "aws_resource_search.handlers.search_resource_handler",
        preview=False,
    )



# -*- coding: utf-8 -*-

"""
This module is an import namespace for everything related to the search.
"""

from .terminal import terminal
from .terminal import format_shortcut
from .terminal import highlight_text
from .terminal import format_resource_type
from .terminal import format_key
from .terminal import format_value
from .terminal import format_key_value
from .terminal import ShortcutEnum
from .terminal import SUBTITLE
from .terminal import SHORT_SUBTITLE
from .searcher_enum import SearcherEnum
from .downloader import T_RESULT_DATA
from .downloader import ResourceIterproxy
from .downloader import ResultPath
from .downloader import list_resources
from .downloader import extract_tags
from .documents.api import BaseArsDocument
from .documents.api import T_ARS_DOCUMENT
from .documents.api import ResourceTypeDocument
from .documents.api import get_utc_now
from .documents.api import to_human_readable_elapsed
from .documents.api import to_utc_dt
from .documents.api import SIMPLE_DT_FMT
from .documents.api import to_simple_dt_fmt
from .documents.api import to_iso_dt_fmt
from .documents.api import get_none_or_default
from .documents.api import get_description
from .documents.api import get_datetime
from .documents.api import get_datetime_simple_fmt
from .documents.api import get_datetime_iso_fmt
from .documents.api import ResourceDocument
from .documents.api import T_ARS_RESOURCE_DOCUMENT
from .items.api import BaseArsItem
from .items.api import T_ARS_ITEM
from .items.api import DetailItem
from .items.api import ExceptionItem
from .items.api import FileItem
from .items.api import InfoItem
from .items.api import UrlItem
from .items.api import AwsResourceTypeItem
from .items.api import AwsResourceItem
from .items.api import SetAwsProfileItem
from .items.api import ShowAwsInfoItem
from .base_searcher import preprocess_query
from .base_searcher import BaseSearcher
from .base_searcher import T_SEARCHER
from .conf.init import config



# -*- coding: utf-8 -*-

from pathlib import Path
from aws_resource_search.items.file_item import FileItem


class TestFileItem:
    def test_from_path(self):
        item = FileItem.from_file(path=Path(__file__), uid="uid")
        assert str(item.variables["path"]) == __file__
        # ------------------------------------------------------------------
        # We only test this action manually on developer's machine, not in CI.
        # ------------------------------------------------------------------
        # item.enter_handler(ui=None)


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(
        __file__,
        "aws_resource_search.items.file_item",
        preview=False,
    )



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import sayt.api as sayt
import aws_arns.api as arns
import aws_console_url.api as acu

from .. import res_lib as rl

if T.TYPE_CHECKING:
    from ..ars_def import ARS


@dataclasses.dataclass
class SqsQueue(rl.ResourceDocument):
    # fmt: off
    queue_arn: str = dataclasses.field(metadata={"field": sayt.StoredField(name="queue_arn")})
    # fmt: on

    @property
    def queue_url(self) -> str:
        return self.raw_data

    @property
    def queue_name(self) -> str:
        q = arns.res.SqsQueue.from_queue_url(url=self.queue_url)
        return q.queue_name

    @property
    def is_fifo(self) -> bool:
        return self.queue_name.endswith(".fifo")

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        q = arns.res.SqsQueue.from_queue_url(url=resource)
        return cls(
            raw_data=resource,
            id=q.queue_name,
            name=q.queue_name,
            queue_arn=q.to_arn(),
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("name", self.name)

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.queue_arn

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.sqs.get_queue(name_or_arn_or_url=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.sqs.queues

    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        detail_items.extend(
            [
                from_detail("queue_url", self.queue_url, url=url),
            ]
        )

        # fmt: off
        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.sqs_client.get_queue_attributes(
                QueueUrl=self.queue_url,
                AttributeNames=["All"],
            )
            attrs = res.get("Attributes", {})
            is_fifo = self.is_fifo
            n_msg = attrs.get("ApproximateNumberOfMessages", "NA")
            n_msg_delayed = attrs.get("ApproximateNumberOfMessagesDelayed", "NA")
            n_msg_invisible = attrs.get("ApproximateNumberOfMessagesNotVisible", "NA")
            policy = attrs.get("Policy")
            redrive_policy = attrs.get("RedrivePolicy")
            redrive_allow_policy = attrs.get("RedriveAllowPolicy")
            detail_items.extend([
                from_detail("is_fifo", is_fifo, url=url),
                from_detail("n_msg", n_msg, url=url),
                from_detail("n_msg_delayed", n_msg_delayed, url=url),
                from_detail("n_msg_invisible", n_msg_invisible, url=url),
                from_detail("policy", policy, value_text=self.one_line(policy), url=url),
                from_detail("redrive_policy", redrive_policy, value_text=self.one_line(redrive_policy), url=url),
                from_detail("redrive_allow_policy", redrive_allow_policy, value_text=self.one_line(redrive_allow_policy), url=url),
            ])
        # fmt: on

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.sqs_client.list_queue_tags(QueueUrl=self.queue_url)
            tags = rl.extract_tags(res)
            detail_items.extend(rl.DetailItem.from_tags(tags, url))

        return detail_items


class SqsQueueSearcher(rl.BaseSearcher[SqsQueue]):
    pass


sqs_queue_searcher = SqsQueueSearcher(
    # list resources
    service="sqs",
    method="list_queues",
    is_paginator=True,
    default_boto_kwargs={
        "PaginationConfig": {
            "MaxItems": 9999,
            "PageSize": 1000,
        },
    },
    result_path=rl.ResultPath("QueueUrls"),
    # extract document
    doc_class=SqsQueue,
    # search
    resource_type=rl.SearcherEnum.sqs_queue.value,
    fields=SqsQueue.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.sqs_queue.value),
    more_cache_key=None,
)



# -*- coding: utf-8 -*-

"""
Key Management Service (KMS) related resources.
"""

import typing as T
import dataclasses

import sayt.api as sayt
import aws_console_url.api as acu

from .. import res_lib as rl

if T.TYPE_CHECKING:
    from ..ars_def import ARS


def normalize_alias(alias: str) -> str:
    if alias.startswith("alias/"):
        return alias[6:]
    else:
        return alias


@dataclasses.dataclass
class KmsKeyAlias(rl.ResourceDocument):
    # fmt: off
    key_id: str = dataclasses.field(metadata={"field": sayt.NgramWordsField(name="key_id", minsize=2, maxsize=4, stored=True)})
    # fmt: on

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id=normalize_alias(resource["AliasName"]),
            name=normalize_alias(resource["AliasName"]),
            key_id=resource.get("TargetKeyId", "NA"),
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("name", self.name)

    @property
    def subtitle(self) -> str:
        return "{}, {}".format(
            rl.format_key_value("key_id", self.key_id),
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.raw_data["AliasArn"]

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.kms.get_key(key_id_or_alias_or_arn=self.key_id)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.kms.customer_managed_keys

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        detail_items.extend([
            from_detail("KeyId", self.raw_data.get("TargetKeyId", "NA"), url=url),
        ])

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.kms_client.describe_key(KeyId=self.key_id)
            dct = res["KeyMetadata"]
            detail_items.extend([
                from_detail("KeyManager", dct.get("KeyManager", "NA"), url=url),
                from_detail("Description", dct.get("Description", "No description"), url=url),
                from_detail("CreationDate", dct.get("CreationDate", "NA"), url=url),
                from_detail("DeletionDate", dct.get("DeletionDate", "NA"), url=url),
                from_detail("ValidTo", dct.get("ValidTo", "NA"), url=url),
                from_detail("KeyState", dct.get("KeyState", "NA"), url=url),
                from_detail("KeyUsage", dct.get("KeyUsage", "NA"), url=url),
            ])

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.kms_client.list_resource_tags(KeyId=self.key_id)
            tags = rl.extract_tags(res)
            detail_items.extend(rl.DetailItem.from_tags(tags, url))
        return detail_items
    # fmt: on


class KmsKeyAliasSearcher(rl.BaseSearcher[KmsKeyAlias]):
    pass


kms_key_alias_searcher = KmsKeyAliasSearcher(
    # list resources
    service="kms",
    method="list_aliases",
    is_paginator=True,
    default_boto_kwargs={
        "PaginationConfig": {"MaxItems": 9999, "PageSize": 1000},
    },
    result_path=rl.ResultPath("Aliases"),
    # extract document
    doc_class=KmsKeyAlias,
    # search
    resource_type=rl.SearcherEnum.kms_key_alias.value,
    fields=KmsKeyAlias.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.kms_key_alias.value),
    more_cache_key=None,
)



# -*- coding: utf-8 -*-

from aws_resource_search.items.url_item import UrlItem


class TestFileItem:
    def test_from_path(self):
        item = UrlItem.from_url(url="https://www.google.com", uid="uid")
        assert str(item.variables["url"]) == "https://www.google.com"
        # ------------------------------------------------------------------
        # We only test this action manually on developer's machine, not in CI.
        # ------------------------------------------------------------------
        # item.enter_handler(ui=None)


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(
        __file__,
        "aws_resource_search.items.url_item",
        preview=False,
    )



# -*- coding: utf-8 -*-

import typing as T
import json
import dataclasses

import aws_console_url.api as acu

from .. import res_lib as rl

if T.TYPE_CHECKING:
    from ..ars_def import ARS


@dataclasses.dataclass
class SecretsManagerSecret(rl.ResourceDocument):
    @property
    def description(self) -> str:
        return rl.get_description(self.raw_data, "Description")

    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        return cls(
            raw_data=resource,
            id=resource["Name"],
            name=resource["Name"],
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("name", self.name)

    @property
    def subtitle(self) -> str:
        return "{}, {}".format(
            self.description,
            self.short_subtitle,
        )

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.raw_data["ARN"]

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.secretmanager.get_secret(secret_name_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.secretmanager.secrets

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_details = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.secretsmanager_client.describe_secret(SecretId=self.name)
            detail_items.extend([
                from_details("KmsKeyId", res.get("KmsKeyId", "NA"), url=url),
                from_details("RotationEnabled", res.get("RotationEnabled", "NA"), url=url),
                from_details("RotationLambdaARN", res.get("RotationLambdaARN", "NA"), url=url),
                from_details("RotationRules", json.dumps(res.get("RotationRules", {})), url=url),
                from_details("LastRotatedDate", res.get("LastRotatedDate", "NA"), url=url),
                from_details("LastChangedDate", res.get("LastChangedDate", "NA"), url=url),
                from_details("LastAccessedDate", res.get("LastAccessedDate", "NA"), url=url),
                from_details("DeletedDate", res.get("DeletedDate", "NA"), url=url),
                from_details("NextRotationDate", res.get("NextRotationDate", "NA"), url=url),
                from_details("OwningService", res.get("OwningService", "NA"), url=url),
                from_details("CreatedDate", res.get("CreatedDate", "NA"), url=url),
                from_details("PrimaryRegion", res.get("PrimaryRegion", "NA"), url=url),
            ])
            tags = rl.extract_tags(res)
            detail_items.extend(rl.DetailItem.from_tags(tags, url))

        return detail_items
    # fmt: on


class SecretsManagerSecretSearcher(rl.BaseSearcher[SecretsManagerSecret]):
    pass


secretsmanager_secret_searcher = SecretsManagerSecretSearcher(
    # list resources
    service="secretsmanager",
    method="list_secrets",
    is_paginator=True,
    default_boto_kwargs={
        "IncludePlannedDeletion": True,
        "SortOrder": "desc",
        "PaginationConfig": {"MaxItems": 9999, "PageSize": 100},
    },
    result_path=rl.ResultPath("SecretList"),
    # extract document
    doc_class=SecretsManagerSecret,
    # search
    resource_type=rl.SearcherEnum.secretsmanager_secret.value,
    fields=SecretsManagerSecret.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(
        rl.SearcherEnum.secretsmanager_secret.value
    ),
    more_cache_key=None,
)



# -*- coding: utf-8 -*-

"""
AWS Resource UI class declaration modules.
"""

import typing as T

import zelfred.api as zf

from . import res_lib as rl
from .handlers.api import (
    open_config_handler,
    search_aws_profile_handler,
    search_resource_type_handler,
    search_resource_handler,
    show_aws_info_handler,
)

if T.TYPE_CHECKING:  # pragma: no cover
    from .ars_def import ARS


def handler(
    query: str,
    ui: "UI",
    skip_ui: bool = False,
) -> T.List[
    T.Union[
        rl.ShowAwsInfoItem,
        rl.SetAwsProfileItem,
        rl.AwsResourceTypeItem,
        rl.AwsResourceItem,
        rl.DetailItem,
        rl.ExceptionItem,
        rl.FileItem,
        rl.InfoItem,
        rl.UrlItem,
    ]
]:  # pragma: no cover
    """
    Main query handler. It parses the query and route the query to
    the corresponding sub handler.

    - :func:`~aws_resource_search.ui.search_resource_type.select_resource_type_handler`
    - :func:`~aws_resource_search.ui.search_resource.search_resource_handler`

    The query can be in one of these formats:

    todo ...

    :param skip_ui: if True, skip the UI related logic, just return the items.
        this argument is used for third party integration.
    """
    zf.debugger.log(f"handler Query: {query!r}")

    # srv id is the service_id-resource_type compound identifier
    # req query is the query string for the resource search
    q = zf.QueryParser(delimiter=":").parse(query)

    # --- handle special commands ---
    # example: s3-bucket: my bucket!?
    if len(query.split("!?", 1)) > 1:
        line_input, _ = query.split("!?", 1)
        return show_aws_info_handler(
            ui=ui,
            line_input=line_input,
            skip_ui=skip_ui,
        )

    # example: s3-bucket: my bucket!@switch to another profile
    if len(query.split("!@", 1)) > 1:
        line_input, aws_profile_query = query.split("!@", 1)
        return search_aws_profile_handler(
            ui=ui,
            line_input=line_input,
            profile_query=aws_profile_query,
            skip_ui=skip_ui,
        )

    # example: s3-bucket: my bucket!{
    if len(query.split("!{", 1)) > 1:
        line_input, sub_query = query.split("!{", 1)
        return open_config_handler(
            ui=ui,
            line_input=line_input,
            skip_ui=skip_ui,
        )

    # example: "  "
    if len(q.trimmed_parts) == 0:
        return search_resource_type_handler(
            ui=ui,
            query="*",
            skip_ui=skip_ui,
        )

    # example:
    # - "ec2 inst"
    # - "s3-bucket"
    # - "ec2 inst: "
    # - "s3-bucket: "
    elif len(q.trimmed_parts) == 1:
        service_query = q.trimmed_parts[0]

        # example:
        # - "ec2 inst: "
        # - "s3-bucket: "
        if len(q.parts) == 1:
            return search_resource_type_handler(
                ui=ui,
                query=service_query,
                skip_ui=skip_ui,
            )

        # example:
        # - "ec2 inst"
        # - "s3-bucket"
        else:
            resource_query = query.split(":")[1].strip()

            # example: "s3-bucket"
            if ui.ars.searcher_finder.is_valid_resource_type(service_query):
                return search_resource_handler(
                    ui=ui,
                    resource_type=service_query,
                    query=resource_query,
                    skip_ui=skip_ui,
                )

            # example: "ec2 inst"
            else:
                return search_resource_type_handler(
                    ui=ui,
                    query=service_query,
                    skip_ui=skip_ui,
                )

    # example: "ec2 inst: something", "s3-bucket: something"
    else:
        # example:
        # - s3-bucket: resource query", "s3-bucket" is a valid srv_id
        # use "resource query" to search
        service_query = q.trimmed_parts[0]
        resource_query = query.split(":")[1].strip()
        if ui.ars.searcher_finder.is_valid_resource_type(service_query):
            return search_resource_handler(
                ui=ui,
                resource_type=service_query,
                query=resource_query,
                skip_ui=skip_ui,
            )

        # example: # ec2 inst: something", "ec2 inst" is not a valid srv_id
        else:
            return search_resource_type_handler(
                ui=ui,
                query=service_query,
                skip_ui=skip_ui,
            )


class UI(zf.UI):
    """
    Extend the ``zelfred.UI`` class to add custom key bindings.
    """

    def __init__(
        self,
        ars: "ARS",
        **kwargs,
    ):
        self.ars: "ARS" = ars
        super().__init__(handler=handler, **kwargs)

    @classmethod
    def new(
        cls,
        ars: "ARS",
    ):
        """
        A factory method to create a new UI instance.
        """
        return UI(
            ars=ars,
            hello_message="Welcome to AWS Resource Search!",
            capture_error=False,
            terminal=rl.terminal,
        )

    def process_ctrl_b(self):  # pragma: no cover
        """
        If you are searching an AWS resource, it will remove the query but keep
        the resource type, so you can enter a new query. For example
        (``|`` is the cursor):

        Now::

            (Query) s3-bucket: my-bucket|

        After Ctrl + B::

            (Query) s3-bucket: |

        If you are searching a sub resource, it will remove the query of the
        sub resource, so you can enter a new query. For example:

        Now::

            (Query) sfn-execution: my-statemachine@a1b2c3|

        After Ctrl + B::

            (Query) sfn-execution: my-statemachine@|
        """
        line = self.line_editor.line
        parts = line.split(":", 1)
        n_parts = len(parts)
        # "sfn-state" -> "sfn-state"
        if n_parts in [0, 1]:
            return
        # "sfn-statemachine:   " -> "sfn-statemachine: "
        elif bool(parts[1].strip()) is False:
            return
        else:
            chunks = parts[1].split("@", 1)
            n_chunks = len(chunks)
            # "sfn-statemachine: my-statemachine" -> "sfn-statemachine: "
            if n_chunks in [0, 1]:
                self.line_editor.clear_line()
                self.line_editor.enter_text(f"{parts[0]}: ")
                return
            # "sfn-statemachine: my-statemachine@  " -> "sfn-statemachine: "
            elif bool(chunks[1].strip()) is False:
                self.line_editor.clear_line()
                self.line_editor.enter_text(f"{parts[0]}: ")
                return
            # "sfn-statemachine: my-statemachine@a1b2c3d4  " -> "sfn-statemachine: my-statemachine@"
            else:
                self.line_editor.clear_line()
                self.line_editor.enter_text(line.split("@", 1)[0] + "@")
                return



# -*- coding: utf-8 -*-

from aws_resource_search.res.s3 import S3Bucket
from aws_resource_search.items.aws_resource_item import AwsResourceItem


class TestAwsResourceItem:
    def test_from_many_document(self):
        item = AwsResourceItem.from_many_document(
            resource_type="s3-bucket",
            docs=[
                S3Bucket(
                    id="my-bucket",
                    name="my-bucket",
                    raw_data={},
                ),
            ],
        )[0]
        _ = item.get_id()
        _ = item.get_name()
        _ = item.get_arn()


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(
        __file__,
        "aws_resource_search.items.aws_resource_item",
        preview=False,
    )



# -*- coding: utf-8 -*-

import zelfred.api as zf

from .ars_init import ars
from .ui_def import UI


ui = UI.new(ars)


def run_ui():
    """
    Run the AWS resource search UI. This is the entry point of the CLI command.
    """
    zf.debugger.reset()
    zf.debugger.enable()
    ui.run()



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import aws_arns.api as arns
import aws_console_url.api as acu

from .. import res_lib as rl

if T.TYPE_CHECKING:
    from ..ars_def import ARS


@dataclasses.dataclass
class SnsTopic(rl.ResourceDocument):
    @classmethod
    def from_resource(cls, resource, bsm, boto_kwargs):
        title = arns.res.SnsTopic.from_arn(resource["TopicArn"]).resource_id
        return cls(
            raw_data=resource,
            id=title,
            name=title,
        )

    @property
    def title(self) -> str:
        return rl.format_key_value("name", self.name)

    @property
    def autocomplete(self) -> str:
        return self.name

    @property
    def arn(self) -> str:
        return self.raw_data["TopicArn"]

    def get_console_url(self, console: acu.AWSConsole) -> str:
        return console.sns.get_topic(name_or_arn=self.arn)

    @classmethod
    def get_list_resources_console_url(cls, console: acu.AWSConsole) -> str:
        return console.sns.topics

    # fmt: off
    def get_details(self, ars: "ARS") -> T.List[rl.DetailItem]:
        from_detail = rl.DetailItem.from_detail
        url = self.get_console_url(console=ars.aws_console)
        detail_items = rl.DetailItem.get_initial_detail_items(doc=self, ars=ars)

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.sns_client.get_topic_attributes(TopicArn=self.arn)
            access_policy = res.get("Attributes", {}).get("Policy")
            delivery_policy = res.get("Attributes", {}).get("DeliveryPolicy")
            subscriptions_confirmed = res.get("Attributes", {}).get("SubscriptionsConfirmed", "NA")
            subscriptions_deleted = res.get("Attributes", {}).get("SubscriptionsDeleted", "NA")
            subscriptions_pending = res.get("Attributes", {}).get("SubscriptionsPending", "NA")
            is_fifo_topic = res.get("Attributes", {}).get("FifoTopic", "NA")
            content_based_deduplication_enabled = res.get("Attributes", {}).get("ContentBasedDeduplication", "NA")
            detail_items.extend([
                from_detail("access_policy", access_policy, value_text=self.one_line(access_policy), url=url),
                from_detail("delivery_policy", delivery_policy, value_text=self.one_line(delivery_policy), url=url),
                from_detail("subscriptions_confirmed", subscriptions_confirmed, url=url),
                from_detail("subscriptions_deleted", subscriptions_deleted, url=url),
                from_detail("subscriptions_pending", subscriptions_pending, url=url),
                from_detail("is_fifo_topic", is_fifo_topic, url=url),
                from_detail("content_based_deduplication_enabled", content_based_deduplication_enabled, url=url),
            ])

        with rl.DetailItem.error_handling(detail_items):
            res = ars.bsm.sns_client.list_tags_for_resource(ResourceArn=self.arn)
            tags = rl.extract_tags(res)
            detail_items.extend(rl.DetailItem.from_tags(tags, url))

        return detail_items
    # fmt: on


class SnsTopicSearcher(rl.BaseSearcher[SnsTopic]):
    pass


sns_topic_searcher = SnsTopicSearcher(
    # list resources
    service="sns",
    method="list_topics",
    is_paginator=True,
    default_boto_kwargs={
        "PaginationConfig": {
            "MaxItems": 9999,
        },
    },
    result_path=rl.ResultPath("Topics"),
    # extract document
    doc_class=SnsTopic,
    # search
    resource_type=rl.SearcherEnum.sns_topic.value,
    fields=SnsTopic.get_dataset_fields(),
    cache_expire=rl.config.get_cache_expire(rl.SearcherEnum.sns_topic.value),
    more_cache_key=None,
)



# -*- coding: utf-8 -*-

"""
A terminal application that enables interactive searches for AWS resources.
It is a mini AWS console in your terminal or shell environment.
"""


from ._version import __version__

__short_description__ = (
    "A terminal application that enables interactive searches for AWS resources. "
    "It is a mini AWS console in your terminal or shell environment."
)
__license__ = "GPL"
__author__ = "Sanhe Hu"
__author_email__ = "husanhe@gmail.com"
__maintainer__ = "Sanhe Hu"
__maintainer_email__ = "husanhe@gmail.com"
__github_username__ = "MacHu-GWU"



# -*- coding: utf-8 -*-

"""
Improve the original enum module.
"""

import typing as T
import enum


class Mixin:
    @classmethod
    def get_by_name(cls, name: str):
        return cls[name]

    @classmethod
    def is_valid_name(cls, name: str) -> bool:
        try:
            _ = cls[name]
            return True
        except KeyError:
            return False

    @classmethod
    def get_names(cls) -> T.List[str]:
        return [i.name for i in cls]


class BetterIntEnum(Mixin, int, enum.Enum):
    """
    Example:

    .. code-block:: python

        >>> class CodeEnum(BetterIntEnum):
        ...     succeeded = 1
        ...     failed = 0
        >>> CodeEnum.get_by_name("succeeded")
        <CodeEnum.succeeded: 1>
        >>> CodeEnum.get_by_value(1)
        <CodeEnum.succeeded: 1>
        >>> CodeEnum.is_valid_name("succeeded")
        True
        >>> CodeEnum.is_valid_name("SUCCEEDED")
        False
        >>> CodeEnum.is_valid_name(1)
        False
        >>> CodeEnum.is_valid_value("succeeded")
        False
        >>> CodeEnum.is_valid_value("SUCCEEDED")
        False
        >>> CodeEnum.is_valid_value(1)
        True
        >>> CodeEnum.ensure_is_valid_value(1)
        >>> CodeEnum.ensure_is_valid_value("succeeded")
        Traceback (most recent call last):
        ...
        ValueError: Invalid CodeEnum: 'succeeded'
        >>> CodeEnum.ensure_int(1)
        1
        >>> CodeEnum.ensure_int(CodeEnum.succeeded)
        1
        >>> isinstance(CodeEnum.ensure_int(1), int)
        True
    """

    @classmethod
    def get_by_value(cls, value: int):
        return cls(value)

    @classmethod
    def is_valid_value(cls, value: int) -> bool:
        try:
            _ = cls(value)
            return True
        except ValueError:
            return False

    @classmethod
    def ensure_is_valid_value(cls, value):
        if cls.is_valid_value(value) is False:
            raise ValueError(f"Invalid {cls.__name__}: {value!r}")

    @classmethod
    def ensure_int(cls, value: T.Union[int, "BetterIntEnum"]) -> int:
        if isinstance(value, cls):
            return value.value
        else:
            return cls(value).value

    @classmethod
    def get_values(cls) -> T.List[int]:
        return [i.value for i in cls]


class BetterStrEnum(Mixin, str, enum.Enum):
    """
    Example:

    .. code-block:: python

        >>> class StatusEnum(BetterStrEnum):
        ...     succeeded = "SUCCEEDED"
        ...     failed = "FAILED"
        >>> StatusEnum.get_by_name("succeeded")
        <StatusEnum.succeeded: 'SUCCEEDED'>
        >>> StatusEnum.get_by_value("SUCCEEDED")
        <StatusEnum.succeeded: 'SUCCEEDED'>
        >>> StatusEnum.is_valid_name("succeeded")
        True
        >>> StatusEnum.is_valid_name("SUCCEEDED")
        False
        >>> StatusEnum.is_valid_value("succeeded")
        False
        >>> StatusEnum.is_valid_value("SUCCEEDED")
        True
        >>> StatusEnum.ensure_is_valid_value("SUCCEEDED")
        >>> StatusEnum.ensure_is_valid_value("succeeded")
        Traceback (most recent call last):
        ...
        ValueError: Invalid StatusEnum: 'succeeded'
        >>> StatusEnum.ensure_str("SUCCEEDED")
        'SUCCEEDED'
        >>> StatusEnum.ensure_str(StatusEnum.succeeded)
        'SUCCEEDED'
        >>> isinstance(StatusEnum.ensure_str("SUCCEEDED"), str)
        True
    """

    @classmethod
    def get_by_value(cls, value: str):
        return cls(value)

    @classmethod
    def is_valid_value(cls, value: str) -> bool:
        try:
            _ = cls(value)
            return True
        except ValueError:
            return False

    @classmethod
    def ensure_is_valid_value(cls, value):
        if cls.is_valid_value(value) is False:
            raise ValueError(f"Invalid {cls.__name__}: {value!r}")

    @classmethod
    def ensure_str(cls, value: T.Union[str, "BetterStrEnum"]) -> str:
        if isinstance(value, cls):
            return value.value
        else:
            return cls(value).value

    @classmethod
    def get_values(cls) -> T.List[str]:
        return [i.value for i in cls]



# -*- coding: utf-8 -*-

from aws_resource_search.items.show_aws_info_item import ShowAwsInfoItem
from aws_resource_search.handlers.show_aws_info_handler import (
    show_aws_info_handler,
)
from aws_resource_search.tests.fake_aws.main import FakeAws


class TestShowAwsInfoHandler(FakeAws):
    @classmethod
    def setup_class_post_hook(cls):
        cls.setup_ars()
        cls.setup_ui()

    def test_show_aws_info_handler(self):
        """
        We don't test handler.
        """
        items = show_aws_info_handler(ui=self.ui, line_input="")


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(
        __file__,
        "aws_resource_search.handlers.show_aws_info_handler",
        preview=False,
    )



# -*- coding: utf-8 -*-

"""
This module provides utilities to download AWS resource data from AWS API.
"""

import typing as T
import dataclasses

import jmespath
from iterproxy import IterProxy

from .base_model import BaseModel

if T.TYPE_CHECKING:  # pragma: no cover
    from boto_session_manager import BotoSesManager
    import sayt.api as sayt

T_RESULT_DATA = T.Union["sayt.T_DOCUMENT", str]
"""
Type hint for boto3 API result data. Each one represents a single AWS resource.
"""


class ResourceIterproxy(IterProxy[T_RESULT_DATA]):
    """
    Advanced iterator object for AWS resource data in boto3 API response.

    Ref: https://github.com/MacHu-GWU/iterproxy-project
    """


@dataclasses.dataclass
class ResultPath(BaseModel):
    """
    Defines how to extract list of AWS resource data from boto3 API call response.

    For example, the
    `s3_client.list_buckets <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/client/list_buckets.html>`_
    API call returns the following response:

    .. code-block:: python

        {
            'Buckets': [
                {
                    'Name': 'string',
                    'CreationDate': datetime(2015, 1, 1)
                },
            ],
            'Owner': {
                'DisplayName': 'string',
                'ID': 'string'
            }
        }

    We aim to extract the list of S3 bucket data from the ``Buckets`` field
    of the response. Similarly, for EC2 Instance, the result path for
    ``ec2_client.describe_instances <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2/client/describe_instances.html>`_
    API call response is ``Reservations[].Instances[]``.

    :param path: the `jmespath <https://jmespath.org/>`_ notation to the result path.
        It will return an empty list if the result path doesn't exist in the response.
    :param _compiled: the compiled jmespath expression. This class will be created
        only once for each AWS resource type, so that we should cache it for
        better performance.
    """

    path: str = dataclasses.field()
    _compiled: jmespath.parser.ParsedResult = dataclasses.field(init=False)

    def __post_init__(self):
        self._compiled = jmespath.compile(self.path + " || `[]`")

    def extract(self, response: dict) -> T.Iterator[T_RESULT_DATA]:
        """
        Extract list of AWS resource data from boto3 API call response.

        :param response: original boto3 API response

        :return: for example, for ``s3_client.list_buckets``, it will return:

        .. code-block:: python

            [
                {
                    'Name': 'string',
                    'CreationDate': datetime(2015, 1, 1)
                },
                ...
            ]
        """
        return self._compiled.search(response)


def list_resources(
    bsm: "BotoSesManager",
    service: str,
    method: str,
    is_paginator: bool,
    boto_kwargs: T.Optional[dict],
    result_path: ResultPath,
) -> ResourceIterproxy:
    """
    Call boto3 API to list AWS resources.

    Example:

    .. code-block:: python

        >>> for iam_group_data in list_resources(
        ...     bsm=bsm,
        ...     service="iam",
        ...     method="list_groups",
        ...     is_paginator=True,
        ...     boto_kwargs=dict(
        ...         PaginationConfig=dict(
        ...             MaxItems=9999,
        ...             PageSize=1000,
        ...         )
        ...     ),
        ...     result_path=ResultPath(path="Groups"),
        ... ):
        ...     print(iam_group_data)

    :param bsm: the ``boto_session_manager.BotoSesManager`` object.
    :param service: the AWS service name for creating the boto3 client.
        for example, the AWS S3 service name is ``s3``.
    :param method: the boto3 client API method to call for listing AWS resources.
        for example, we use ``list_buckets`` method for getting AWS S3 buckets,
        we use ``describe_instances`` method for getting AWS EC2 instances.
    :param is_paginator: boolean value to indicate whether the method is a paginator.
        for example, it is False for ``s3.list_buckets`` method,
        it is True for ``ec2.describe_instances`` method.
    :param boto_kwargs: the keyword arguments for the boto3 client API call.
        if it is a paginator, it often contains ``PaginationConfig`` key.
    :param result_path: the :class:`ResultPath` object to extract list of AWS resource
    """

    def func():
        if boto_kwargs is None:
            kwargs = {}
        else:
            kwargs = boto_kwargs
        client = bsm.get_client(service)
        if is_paginator:
            paginator = client.get_paginator(method)
            for response in paginator.paginate(**kwargs):
                yield from result_path.extract(response)
        else:
            response = getattr(client, method)(**kwargs)
            yield from result_path.extract(response)

    return ResourceIterproxy(func())


def extract_tags(data: dict) -> T.Dict[str, str]:
    """
    Extract tags key value pair from boto3 API call response data.

    :param data: it is the dictionary representation of one AWS resource,
        it could be the original boto3 API response, it also could be
        a nested dictionary. For example: in ``s3.list_buckets`` response,
        the original response is the data, because it has a ``TagSet`` field.
        in ``iam.list_roles`` response, the dict in the ``Roles`` list is the data.
    """
    if "tags" in data:
        tag_data = data["tags"]
    elif "Tags" in data:
        tag_data = data["Tags"]
    elif "TagSet" in data:
        tag_data = data["TagSet"]
    elif "TagList" in data:
        tag_data = data["TagList"]
    else:
        return {}

    if isinstance(tag_data, dict):
        return tag_data
    elif isinstance(tag_data, list):
        tags = {}
        for dct in tag_data:
            if "Key" in dct:
                tags[dct["Key"]] = dct["Value"]
            elif "key" in dct:
                tags[dct["key"]] = dct["value"]
            elif "TagKey" in dct:
                tags[dct["TagKey"]] = dct["TagValue"]
            else:
                raise ValueError(f"unable to extract tags from {data}")
        return tags
    else:
        raise TypeError(f"unable to extract tags from {data}")



# -*- coding: utf-8 -*-

from aws_resource_search.items.info_item import InfoItem


class TestInfoItem:
    def test(self):
        _ = InfoItem


if __name__ == "__main__":
    from aws_resource_search.tests.helper import run_cov_test

    run_cov_test(
        __file__,
        "aws_resource_search.items.info_item",
        preview=False,
    )



# -*- coding: utf-8 -*-

"""
This module manages the singleton object ``ars`` which holds all context data
such as boto session. It also the entry point to access the aws resource search
functionality.
"""

from .ars_def import ARS

ars = ARS.from_profile()
bsm = ars.bsm
